# Libraries
import collections
import pandas as pd

# Import specific
from itertools import product

# -----------------------------------------
# Helper methods
# -----------------------------------------
def length_exceptions(x):
    """This method..."""
    if len(x) == 1:
        return [len(x[0])]
    if len(x) >= 4:
        return [1] * len(x)
    return None

def _loops_strategy_lengths(series):
    """
    :param x:
    :return:
    """
    import numpy as np
    # Create DataFrame with words
    words = series.str.split(expand=True, n=1)
    # Create DataFrame with word lengths
    lengths = words.astype('str').applymap(lambda x: len(x))
    # Create loops
    return list(product(*[range(4, n) for n in lengths.max()]))

# -----------------------------------------
# Constants
# -----------------------------------------
ACRONYM_EXCEPTIONS = [
    length_exceptions
]


def _acronym(x, lengths=None, sep='', exceptions=ACRONYM_EXCEPTIONS):
    """Create an acronym from a string single.

    .. note: Add variable to chose whether we want to keep
             the whole word if the split length is 1.

    Parameters
    ----------
    x: String
        The strings to create the acronym/code.

    sep: String (default ' ')
        The separator to include between the acronym components. For
        instance, the sep values ' ' and '_' for the string
        'artificial intelligence' would lead to 'AI' and 'A_I'
        respectively.

    lengths: tuple
        The number of letters to use from each word after
        the initial string has been split.

    exceptions: dictionary
        The condition to evaluate as key and the lengths to use as values.
        Only the first condition that returns Tue will be valuated. The
        signature of the methods are as follows:

        key function:
        :param x: list - array obtained from split
        :return: boolean - whether

        value function:
        :param x: list - array obtained from split
        :return: list - array with lenghts

    Returns
    --------
    string
        The acronym
    """
    # Splits
    split = x.split()

    # Define default lengths
    if lengths is None:
        lengths = [1] * len(split)

    # Exceptions
    #for k,v in exceptions.items():
    #    if k(split):
    #        lengths = v(split)
    for expt in exceptions:
        aux = expt(split)
        if aux is not None:
            lengths = aux

    # Compute acronym
    return sep.join([c[:l].upper() \
        for c, l in zip(split, lengths)])



def acronym_series_unique(series, split_n=1,
                          exclude_acronyms=[],
                          loops_strategy=None,
                          verbose=10,
                          kwgs_acronym={}):
    """Computes unique acronyms.

    Parameters
    ----------

    Returns
    --------
    """

    # Set default none acronym
    acronym = \
        pd.Series(index=series.index, data=pd.NA, name='acronym')

    # Define loops strategy
    if loops_strategy is None:
        loops_strategy = _loops_strategy_lengths

    # Find loops
    loops = loops_strategy(series)

    # Loop
    for i, l in enumerate(loops):

        # Find duplicates or empty values
        idxs = acronym.duplicated(keep='first') | acronym.isna()

        # Break clause
        if idxs.sum() == 0:
            break

        # Show information
        if verbose > 5:
            print("%s/%s. lengths=%s" % (i, len(loops), l))
        if verbose > 7:
            df = pd.concat([series[idxs], acronym[idxs]], axis=1)
            print("%s\n" % df.sort_values(by='acronym'))

        # Create acronyms
        aux = series[idxs].apply(_acronym, lengths=l, **kwgs_acronym) # Create new
        aux[aux.isin(exclude_acronyms)] = pd.NA                       # Exclude
        acronym[idxs] = aux[idxs]                                     # Update

    # Show warning
    if verbose > 0:
        print("There are %s repeated acronyms!\n" % idxs.sum())

    # Return
    return acronym


def acronym_series(series, unique_acronyms=False, **kwargs):
    """This method...

    Parameters
    ----------
    series: pd.Series
        The series with the names to convert in acronyms.
    unique_acronyms:

    exclude_acronyms:

    split_n: int

    verbose: int
        Level of verbosity

    loops_strategy: function
        The function to indicate what length combinations should
        be used on each iteration. By default it will use the
        default method _loops_strategy_lengths which split the
        series in two and returns all possible lengths from (4, 4)
        till (max_len, max_len). The signature of the function to
        pass as loops_strategy is as follows:

        :param x: series
        :return: list (array of lengths)

    kwgs_split: dict
        The parameters to pass to the split function

    kwgs_acronym: dict
        The parameters to pass to the acronym function.

    Returns
    -------
    pd.Series
        The acronym series

    """
    # Basic check
    if series.duplicated().any() and unique_acronyms:

        # show warning
        repeated = series.value_counts()
        print("\nThe series has the following identical values and \n"
              "therefore they cannot or shouldn't be expressed with \n"
              "different acronyms. The unique_acronyms parameter has \n"
              "been set to 'False'.\n\n{0}" \
                .format(repeated[repeated > 1]))
        # Set unique acronyms to false
        unique_acronyms = False

    # Return acronyms
    if not unique_acronyms:
        return series.apply(_acronym)

    # Return
    return acronym_series_unique(series, **kwargs)


def invert(d):
    return {v:k for k,v in d.items()}


def create_registry(data, keyword=None, keep=None):
    """Creates registry from data.

    Parameters
    ----------
    data: pd.DataFrame
        The data

    keyword: string
        The keyword for the columns. All columns starting
        with such keyword will be kept and used for the
        registry.

    keep: list
        The list of columns to keep for the registry
    """
    # Columns to keep
    if keep is None:
        keep = [c for c in data.columns
            if c.startswith(keyword) and
                (c.endswith('name') or
                 c.endswith('code') or
                 c.endswith('description'))]

    # Copy data
    reg = data[keep].copy(deep=True)
    reg = reg.drop_duplicates()
    reg = reg.reset_index(drop=True)

    # Add id
    #if keyword is not None:
    #    reg['%s_id' % keyword] = reg.index.values

    # Return
    return reg


# -----------------------------------------------------
# Microorganism
# -----------------------------------------------------
# The regexp map
REGEX_MAP_MICROORGANISM = {
    '\([^)]*\)': '',      # Remove everything between ()
    '(\s)*\-(\s)*': '-',  # Remove spaces before after hyphen
    'species': '',        # Rename species for next regexp
    'o157': '',           # Remove scherichia coli o157
    'sp(\.)*(\s|$)+': ' ',  # Remove sp from word.
    'strep(\.|\s|$)': 'streptococcus ',   # Complete
    'staph(\.|\s|$)': 'staphylococcus ',  # Complete
    'staphylococci': 'staphylococcus',    # Correction (add mixed? group?)
    'streptococci': 'streptococcus',      # Correction (add mixed? group?)
    '\s+': ' ',           # Remove duplicated spaces.
}

# The hyphens
HYPHENS = ['haemolytic']

# The words to move to the beginning
MOVE_TO_START = ['enterococcus',
                 'staphylococcus',
                 'streptococcus',
                 'coliform']

# The words to move to the end
MOVE_TO_END = ['methicillin resistant',
               'vancomycin resistant',
               'mixed']


def _clean_microorganism(series,
                         hyphens=HYPHENS,
                         move_to_start=MOVE_TO_START,
                         move_to_end=MOVE_TO_END):
    """Cleans the microorganism names.

    .. todo: Put everything below as our own defined function
             but allow users to pass their own functions.

    Parameters
    ----------
    series: pd.Series
        The series with the name of the organisms. Ideally it
        should represent the binomial nomenclature including
        genus and specie respectively.

    hyphens:

    move_to_start:

    move_to_end:

    Returns
    -------
    """

    # Libraries
    from pyamr.datasets.clean import hyphen_before
    from pyamr.datasets.clean import word_to_start

    # Copy
    s = series.copy(deep=True)

    # Lower
    s = s.str.lower()

    # Apply regex mapping
    s = s.replace(regex=REGEX_MAP_MICROORGANISM)

    # Correct hyphens
    for hp in hyphens:
        s = s.transform(hyphen_before, w=hp)

    # Correct order of genus
    for sp in move_to_start:
        s = s.transform(word_to_start, w=sp, pos='start')

    # Correct order of tags
    for sp in move_to_end:
        s = s.transform(word_to_start, w=sp, pos='end')

    # Apply regexp mapping
    s = s.replace(regex=REGEX_MAP_MICROORGANISM)

    # Final strip
    s = s.str.strip()

    # Return
    return s


def clean_specimen(series):
    """
    \sr\s = right -> end
    \sl\s = left -> end


    :param series:
    :return:
    """

# ---------------------------------------------------
# Registry Base
# ---------------------------------------------------
class Registry:
    """This is basically a lookup table."""

    # The order of the columns within the registry
    ORDER = ['id', 'name', 'code', 'description', 'original']
    # The subset to use to drop duplicates
    SUBSET = ['name', 'original']
    # The function to clean the names
    FCLEAN = {}
    # The dictionary to rename columns
    RENAME_COLUMNS = {}
    # Registry dataframe
    REG = None


    def __init__(self, keyword='',
                       order=ORDER,
                       subset=SUBSET,
                       fclean=FCLEAN):
        """Constructor

        .. note: Raise type errors
        .. note: Allow to load from file
        """
        # Set parameters
        self.keyword = keyword
        self.ORDER = order
        self.SUBSET = subset
        self.FCLEAN = fclean

        # Add columns to order. Note that if they are
        # important to drop duplicates, they probably
        # should be kept.
        #if subset is not None:
        #    self.ORDER += sorted(set(subset).difference(set(order)))


    def getr(self, prepend=False):
        """Returns the registry DataFrame"""
        # Get registry
        aux = self.REG.copy(deep=True)

        # Prepend
        if prepend and self.keyword!='':
            aux.columns = ['%s_%s' % (self.keyword, c)
                for c in aux.columns]

        # Fill na with '' so null=False.
        # aux = aux.fillna(aux.dtypes.replace({'O': ''}))

        # Return
        return aux


    def fit(self, data):
        """This method...

        .. note: It assumes name exists...

        Parameters
        ----------
        data: pd.DataFrame
            The DataFrame expects to have the code, the
            name and the description. Specially the name.
            Think what happens if other missing.

        Returns
        -------
        """
        # Create registry
        aux = create_registry(data, keyword=self.keyword)

        # Keep only last label
        aux.columns = [c.split("_", 1)[-1] for c in aux.columns]

        # Raise error
        if not set(['name', 'code']).intersection(aux.columns):
           raise ValueError("Missing either name or code")

        # Add missing columns
        if 'code' not in aux:
            aux['code'] = aux.name
        if 'name' not in aux:
            aux['name'] = aux.code

        # Backup original
        original = aux.name

        # Put everything to lowercase
        aux = aux.applymap(lambda x: x.lower()
            if isinstance(x, str) else x)

        # Restore original
        aux['original'] = original

        """
        # Add code and fill empty name
        aux.name.fillna(aux.code, inplace=True)

        # Replace
        aux.code = aux.code.replace(self.CODE_REPLACE)

        # Map
        aux = aux.replace({
            'name': self.CODE_MAP,
            'code': invert(self.CODE_MAP)
        })
        """

        print("=====> %s" % self.FCLEAN)

        # Apply cleaning
        for k,v in self.FCLEAN.items():
            if not k in aux.columns:
                continue
            if not callable(v):
                continue
            aux[k] = v(aux[k])
            print("cleaned %s with %s" % (k, v))

        # Final formatting
        aux = aux.drop_duplicates(subset=self.SUBSET)
        aux.name = aux.name.astype(str)
        aux = aux.sort_values(by='name')
        aux = aux.reset_index()
        aux['id'] = aux.index + 1

        # Keep
        keep = [c for c in self.ORDER if c in aux.columns]

        # Order information
        aux = aux[keep]

        # Set registry
        self.REG = aux

        # Return
        return self

    def replace(self, series, key='original', value='name'):
        """This method..."""
        tup = zip(self.REG[key], self.REG[value])
        return series.map(dict(tup))

    def transform(self, data, replace={}, include_id=True):
        """Transform data

        Parameters
        ----------
        data: pd.DataFrame
            The data to transform.
        replace:

        include_id

        Returns
        -------
        pd.DataFrame
            The data transformed
        """
        # Include the id.
        if include_id:
            data['%s_id' % self.keyword] = \
                self.replace(data['%s_name' % self.keyword],
                    key='name', value='id')

        # Perform replace
        #for column, (key, value) in replace.items():
        #    data[column] = self.replace(data[column],
        #        key='original', value='name')

        # Return
        return data


    def fit_transform(self, data, **kwargs):
        """Fits and transforms"""
        self.fit(data)
        return self.transform(data, **kwargs)

    def combine(self, data):
        pass

    def clean(self, series):
        pass


# ---------------------------------------------------
# Registry Microorganism
# ---------------------------------------------------
class MicroorganismRegistry(Registry):
    """Registry for microorganisms"""

    taxonomy = ['domain',
                'class',
                'order',
                'family',
                'genus',
                'species',
                'subspecies']

    reg = None

    def __init__(self, **kwargs):
        """"""
        # Super
        super().__init__(**kwargs)

        # For some reason setting the dictionary as an
        # attribute does not work, the variable FCLEAN
        # remains {} (from parent).
        # Set cleaning dictionary
        self.FCLEAN = {
            'name': _clean_microorganism
        }


    def combine(self, dataframe, on='name'):
        """Combines an external dataframe with the registry.

        .. note: The dataframe must contain genus, species..

        Parameters
        ----------

        Returns
        --------
        """
        # Load
        from pyamr.datasets.load import load_registry_microorganisms

        # Copy DataFrame
        aux = dataframe.copy(deep=True)

        # Create genus and species
        aux[['genus', 'species']] = \
            aux[on] \
                .str.capitalize() \
                .str.split(expand=True, n=1)

        # Format
        aux.genus = aux.genus.str.capitalize()
        aux.species = aux.species.str.lower()

        # Load registry information
        if self.reg is None:
            self.reg = load_registry_microorganisms()

        # --------------
        # Step 1
        # --------------
        # First, merge those rows within gram_stain and taxonomy
        # that have equal genus and species. Note that we are
        # only merging if both values exist and are equal.
        # Merge
        aux = pd.merge(aux, self.reg,
            how='left',
            left_on=['genus', 'species'],
            right_on=['genus', 'species']
        )

        # Those merged exist in registry.
        aux['exists_in_registry'] = aux.acronym.notna()

        # --------------
        # Step 2
        # --------------
        # Second, for those values whose taxonomy-related columns
        # are null, use the taxonomy information based only on the
        # genus. This does not overwrite step 1.
        # Create aux
        aux_step2 = pd.merge(aux[['genus']],
            self.reg.drop(columns=['species', 'acronym']) \
                .drop_duplicates() \
                .groupby('genus') \
                .head(1),
            how='left',
            left_on=['genus'],
            right_on=['genus']
        )

        # Update dataframe
        aux.update(aux_step2)

        # Return
        return aux

    def binomial_name(self):
        pass

    def uuid(self):
        pass


# ---------------------------------------------------
# Registry Antimicrobial
# ---------------------------------------------------
class AntimicrobialRegistry(Registry):
    """Registry for antimicrobials"""

    reg = None

    def combine(self, dataframe, on='name'):
        """Combines an external dataframe with the registry.

        .. note: I am assuming the columns that exist in dataframe...

        Parameters
        ----------

        Returns
        --------
        """
        # Load
        from pyamr.datasets.load import load_registry_antimicrobials

        # Copy DataFrame
        aux = dataframe.copy(deep=True)

        # Format
        aux[on] = aux[on].str.capitalize()

        # Load registry information
        if self.reg is None:
            self.reg = load_registry_antimicrobials()

        # --------------
        # Step 1
        # --------------
        # First, merge those rows within gram_stain and taxonomy
        # that have equal genus and species. Note that we are
        # only merging if both values exist and are equal.
        # Merge
        aux = pd.merge(aux, self.reg,
            how='left',
            left_on=[on],
            right_on=['name']
        )

        # Return
        return aux








if __name__ == '__main__':

    # Libraries
    import pandas as pd

    def len1(x):
        """
        :param x: array obtained from split
        :return: boolean
        """
        return len(x) == 1

    def wordl(x):
        """
        :param x: array obtained from split
        :return: array with lenghts
        """""
        return [len(x[0])]

    # Create data.
    organisms = ['Pseudomonas Aeruginosa',
                 'Pseudomonas Aeruginose', # minor variant
                 'Pseudomonas Aeruginosi', # minor variant
                 'Pseudomonas',            # genus 1
                 'Enterococcus',           # genus 2
                 'Enterococcus vagus',
                 'Staphylococcus Beta-Haemolytic Group A',
                 'This is another example',  # Long
                 'Pseu Aeru',                # Extreme
                 'Enterococcus vagus',       # repeated
                 'Pseudomonas',              # repeated
    ]

    # Create series
    series = pd.Series(organisms)

    # Create new acronyms
    acronyms_1 = series.apply(_acronym, sep='')
    acronyms_2 = series.apply(_acronym, sep='_')

    # Create acronyms
    acronyms_3 = acronym_series(series,
        exclude_acronyms=[],
        unique_acronyms=True,
        verbose=0)

    # -----------------------------
    # Creating unique acronyms
    # -----------------------------
    # Create loops from default loop_strategy
    loops = _loops_strategy_lengths(series)

    # Create acronyms
    acronyms_4 =  acronym_series(series[:-3],
        exclude_acronyms=['PSEU_AERU'],
        unique_acronyms=True,
        verbose=10,
        kwgs_acronym={
            'sep': '_'
        })

    # Create result
    result = pd.DataFrame()
    result['series'] = series
    result['acronyms_1'] = acronyms_1
    result['acronyms_2'] = acronyms_2
    result['acronyms_3'] = acronyms_3
    result['acronyms_4'] = acronyms_4

    # Show
    print("Results:")
    print(result)
    print("Loop Strategy:")
    print(pd.DataFrame(loops))


# Libraries
import warnings
import pandas as pd

# Own
from pyamr.core.sari import SARI

class SART:

    # Attributes
    c_spe = 'SPECIMEN'
    c_org = 'MICROORGANISM'
    c_abx = 'ANTIMICROBIAL'
    c_dat = 'DATE'
    c_out = 'SENSITIVITY'
    c_res = 'RESISTANCE'

    def __init__(self, column_specimen=c_spe,
                       column_microorganism=c_org,
                       column_antimicrobial=c_abx,
                       column_date=c_dat,
                       column_outcome=c_out,
                       column_resistance=c_res):
        """Constructor.

        Parameters
        ----------
        groupby: list
            The labels of the columns to groupby.

        Returns
        --------
        SARI instance
        """
        # Columns
        self.c_spe = column_specimen
        self.c_org = column_microorganism
        self.c_abx = column_antimicrobial
        self.c_dat = column_date
        self.c_out = column_outcome
        self.c_res = column_resistance

        # Groupby
        self.groupby = [self.c_spe,
                        self.c_org,
                        self.c_abx,
                        self.c_out]

    def compute(self, dataframe, period='180D', shift='30D', cdate=None,
                return_objects=True, **kwargs):
        """Computes single antibiotic resistance index.

        .. todo: Add parameters to rolling!
        .. todo: Place value at the left, center, right of window?
        .. todo: Ensure that works when time gaps present!
        .. todo: Carefull with various indexes!

        Parameters
        ----------
        dataframe: pd.DataFrame
            It might receive two different types of DataFrames.

            The first option is a DataFrame with the raw susceptibility test
            records where the interpretation ('sensitive', 'intermediate',
            'resistant') that will be used to compute the sari.

            The second option is a DataFrame with the sari values already
            computed.

        shift: str
            Frequency value to pass to pd.Grouper.

        period: str, int
            Window value to pass to pd.rolling.

        cdate: string, default=None
            The column that will be used as date.

        return_frequencies: boolean, default=True
            Whether to return the frequencies or just the resistance index.

        strategy: string or func, default='hard'
            The method used to compute sari. The possible options
            are 'soft', 'medium' and 'hard'. In addition, a function
            with the following signature func(DataFrame, **kwargs)
            can be passed.

                (i) ``soft``    as R / R+I+S
                (ii) ``medium`` as R / R+S
                (iii) ``hard``  as R+I / R+I+S
                (iv) ``other``  as R+0.5I / R+0.5I+S [Not yet]

        **kwargs: arguments to pass the strategy function.

        Returns
        -------
        pd.Series or pd.DataFrame
            The resistance index (pd.Series) or a pd.DataFrame with the
            resistance index (sari) and the frequencies.
        """
        # Libraries
        import numpy as np
        import pandas as pd
        import statsmodels.api as sm

        # Libraries
        from pyamr.core.sari import SARI
        from pyamr.core.regression.wls import WLSWrapper
        from pyamr.metrics.weights import SigmoidA

        # Copy DataFrame
        aux = dataframe.copy(deep=True)

        # ------------------------------
        # Compute resistance time-series
        # ------------------------------
        # Create SARI instance
        sar = SARI(groupby=self.groupby)

        # Compute sari time-series
        sari_oti = sar.compute(aux, shift=shift,
            period=period, cdate='date_received')

        sari_oti = sari_oti.reset_index()

        # ------------------------
        # Compute resistance trend
        # ------------------------
        # Group by tuples
        groups = sari_oti.groupby(self.groupby[:-1])

        # Save
        objs = []

        # Loop
        for i, (name, group) in enumerate(groups):
            print("%2s/%2s. Computing... %s" % (i+1, len(groups), str(name)))

            # Warning
            if aux.shape[0] < 2:
                warnings.warn("""
                     {}:
                     There is only one single point and therefore the
                     resistance trend cannot be estimated. To generate a more
                     granular time series please reduce the value of <shift> 
                     or alternative find a more complete dataset.\n""".format(name))
                continue

            # Interpolate (if missing dates)
            aux = group \
                .set_index('date_received') \
                .resample(shift) \
                .interpolate(method='linear')

            # Extract variables
            x = np.arange(len(aux.sari.values))
            y = aux.sari.values * 100
            f = aux.freq.values

            # Create method to compute weights from frequencies
            W = SigmoidA(r=200, g=0.5, offset=0.0, scale=1.0)

            # Note that the function fit will call M.weights(weights) inside and will
            # store the M converter in the instance. Therefore, the code executed is
            # equivalent to <weights=M.weights(f)> with the only difference being that
            # the weight converter is not saved.
            wls = WLSWrapper(estimator=sm.WLS).fit( \
                exog=x, endog=y, trend='c', weights=f,
                W=W, missing='raise')

            # Add pearson correlation coefficient
            wls._result['pearson'] = group.sari.corr(group.freq)

            # Append object
            objs += [(name, wls)]

        # Construct DataFrame
        table = pd.DataFrame([
            obj.as_series().append(
                pd.Series({
                    self.c_spe: name[0],
                    self.c_org: name[1],
                    self.c_abx: name[2]
                })
            ) for name, obj in objs])

        # Set index information
        table = table.set_index([
            self.c_spe,
            self.c_org,
            self.c_abx])

        # Return
        if return_objects:
            return table, objs
        return table



if __name__ == '__main__': # pragma: no cover

    # Libraries
    import numpy as np
    import pandas as pd
    import seaborn as sns
    import matplotlib as mpl
    import matplotlib.pyplot as plt

    # Import own libraries
    from pyamr.datasets.load import load_data_nhs

    # -------------------------
    # Configuration
    # -------------------------
    # Configure seaborn style (context=talk)
    sns.set(style="white")

    # Set matplotlib
    mpl.rcParams['xtick.labelsize'] = 9
    mpl.rcParams['ytick.labelsize'] = 9
    mpl.rcParams['axes.titlesize'] = 11
    mpl.rcParams['legend.fontsize'] = 9

    # Pandas configuration
    pd.set_option('display.max_colwidth', 40)
    pd.set_option('display.width', 300)
    pd.set_option('display.precision', 4)

    # Numpy configuration
    np.set_printoptions(precision=2)

    # -------------------------------------------
    # Load data
    # -------------------------------------------
    # Load data
    data, antimicrobials, microorganisms = load_data_nhs()

    # Show
    print("\nData:")
    print(data)
    print("\nColumns:")
    print(data.columns)
    print("\nDtypes:")
    print(data.dtypes)

    # Select tuples
    spec = ['URICUL']
    orgs = ['ECOL']
    abxs = ['ACELX', 'ACIP', 'AAMPC', 'ATRI', 'AAUG',
            'AMER', 'ANIT', 'AAMI', 'ACTX', 'ATAZ',
            'AGEN', 'AERT', 'ACAZ', 'AMEC', 'ACXT']

    # Filter
    idxs_spec = data.specimen_code.isin(spec)
    idxs_orgs = data.microorganism_code.isin(orgs)
    idxs_abxs = data.antimicrobial_code.isin(abxs)

    # Filter
    data = data[idxs_spec & idxs_orgs & idxs_abxs]

    # Filter dates (2016-2018 missing)
    data = data[data.date_received.between('2008-01-01', '2016-12-31')]


    from pyamr.core.sart import SART

    # Create instance
    sar = SART(column_specimen='specimen_code',
               column_microorganism='microorganism_code',
               column_antimicrobial='antimicrobial_code',
               column_date='date_received',
               column_outcome='sensitivity',
               column_resistance='sari')

    # Compute resistance trends
    table = sar.compute(data, shift='30D', period='180D')

    print(table)

    import sys
    sys.exit()

    # -------------------------------------------
    # Compute OTI sari (temporal)
    # -------------------------------------------
    # Generic
    import statsmodels.api as sm
    import statsmodels.robust.norms as norms

    # Libraries
    from pyamr.core.sari import SARI
    from pyamr.core.sart import SART
    from pyamr.datasets.load import make_timeseries
    from pyamr.core.regression.wls import WLSWrapper
    from pyamr.metrics.weights import SigmoidA

    # Variables
    shift, period = '30D', '180D'

    # Create SARI instance
    sar = SARI(groupby=['specimen_code',
                        'microorganism_code',
                        'antimicrobial_code',
                        'sensitivity'])

    # Compute SARI overall
    sari_overall = sar.compute(data,
         return_frequencies=True)

    # Compute sari time-series
    oti = sar.compute(data, shift=shift,
         period=period, cdate='date_received')

    # Show
    print("\nSARI:")
    print(oti)


    # -------------------------------------------
    # Compute resistance trend
    # -------------------------------------------
    # .. note: Be careful that there might be intervals (e.g. months)
    #          which do not appear in the DataFrame because there was
    #          no data (freq=0) to compute the resistance (SARI).

    #sart = SART(column_specimen='specimen_code',
    #            column_microorganism='microorganism_name',
    #            column_antimicrobial='antimicrobial_name',
    #            column_outcome='sensitivity',
    #            column_resistance='sari')#



    #import sys
    #sys.exit()


    # Group by tuples
    groups = oti.reset_index()\
        .groupby(['specimen_code',
                  'microorganism_code',
                  'antimicrobial_code'])

    # Save
    objs = []

    # Loop
    for i, (name, group) in enumerate(groups):
        print("%2s/%2s. Computing... %s" % (i, len(groups), str(name)))

        # Interpolate (if missing dates)
        aux = group \
            .set_index('date_received') \
            .resample(shift) \
            .interpolate(method='linear')

        # Extract variables
        x = np.arange(len(aux.sari.values))
        y = aux.sari.values * 100
        f = aux.freq.values

        # Create method to compute weights from frequencies
        W = SigmoidA(r=200, g=0.5, offset=0.0, scale=1.0)

        # Note that the function fit will call M.weights(weights) inside and will
        # store the M converter in the instance. Therefore, the code executed is
        # equivalent to <weights=M.weights(f)> with the only difference being that
        # the weight converter is not saved.
        wls = WLSWrapper(estimator=sm.WLS).fit( \
            exog=x, endog=y, trend='c', weights=f,
            W=W, missing='raise')

        # Add pearson correlation coefficient
        wls._result['pearson'] = group.sari.corr(group.freq)

        # Display
        # This example shows how to make predictions using the wrapper and how
        # to plot the result in data. In addition, it compares the intervals
        # provided by get_prediction (confidence intervals) and the intervals
        # provided by wls_prediction_std (prediction intervals).
        #
        # To Do: Implement methods to compute CI and PI (see regression).

        # Variables.
        start, end = None, 230

        # Compute predictions (exogenous?). It returns a 2D array
        # where the rows contain the time (t), the mean, the lower
        # and upper confidence (or prediction?) interval.
        preds = wls.get_prediction(start=start, end=end)

        # Create figure
        fig, ax = plt.subplots(1, 1, figsize=(11, 5))

        # Plotting confidence intervals
        # -----------------------------
        # Plot truth values.
        ax.plot(x, y, color='#A6CEE3', alpha=0.5, marker='o',
                markeredgecolor='k', markeredgewidth=0.5,
                markersize=5, linewidth=0.75, label='Observed')

        # Plot forecasted values.
        ax.plot(preds[0, :], preds[1, :], color='#FF0000', alpha=1.00,
                linewidth=2.0, label=wls._identifier(short=True))

        # Plot the confidence intervals.
        ax.fill_between(preds[0, :], preds[2, :],
                        preds[3, :],
                        color='r',
                        alpha=0.1)

        # Legend
        plt.legend()
        plt.title(name)

        # Append object
        objs += [(name, wls)]




    # -------------------------------------------
    # Plotting Table Graph
    # -------------------------------------------
    # Libraries
    from pyamr.graphics.table_graph import _DEFAULT_CONFIGURATION
    from pyamr.graphics.table_graph import vlinebgplot

    # Configuration
    info = _DEFAULT_CONFIGURATION

    rename = {
        'wls-x1_coef': 'sart_m',
        #'wls-const_coef': 'offset',
        'wls-rsquared': 'r2',
        'wls-rsquared_adj': 'r2_adj',
        'wls-m_skew': 'skew',
        'wls-m_kurtosis': 'kurtosis',
        'wls-m_jb_prob': 'jb',
        'wls-m_dw': 'dw',
        'wls-const_tprob': 'ptm',
        'wls-x1_tprob': 'ptn',
        'wls-pearson': 'pearson',
        'freq': 'freq',
    }

    # Construct DataFrame
    table = pd.DataFrame([
        obj.as_series().append(
            pd.Series({
                'specimen_code': name[0],
                'microorganism_code': name[1],
                'antimicrobial_code': name[2]
            })
        ) for name, obj in objs])

    # Set index information
    table = table.set_index([
        'specimen_code',
        'microorganism_code',
        'antimicrobial_code'
    ])

    # Combine and format
    comb = table.join(sari_overall)
    comb.index = comb.index.map('_'.join)
    comb = comb.reset_index()
    comb = comb.rename(columns=rename)
    comb['sart_y'] = comb.sart_m * 12   # Yearly trend
    comb['sari_pct'] = comb.sari * 100  # SARI percent

    # Sort by trend
    comb = comb.sort_values(by='sart_y', ascending=False)

    # Select only numeric columns
    #data = comb.select_dtypes(include=np.number)
    data = comb[[
        'index',
        'sart_m',
        'sart_y',
        'sari_pct',
        'r2',
        'r2_adj',
        'skew',
        'kurtosis',
        'jb',
        'dw',
        'ptm',
        'ptn',
        'pearson',
        'freq'
    ]]

    # Show DataFrame
    print("\nResults:")
    print(data)

    # Create pair grid
    g = sns.PairGrid(data, x_vars=data.columns[1:],
        y_vars=["index"], height=3, aspect=.45)

    # Draw a dot plot using the stripplot function
    #g.map(sns.stripplot, size=10, orient="h", jitter=False,
    #      palette="flare_r", linewidth=1, edgecolor="w")

    # Set common features
    g.set(xlabel='', ylabel='')

    # Plot strips and format axes (skipping index)
    for ax, c in zip(g.axes.flat, data.columns[1:]):

        # Get information
        d = info[c] if c in info else {}

        # Display
        # sns.stripplot(data=data, x=title, y='abx', size=10,
        #    orient="h", jitter=False, linewidth=0.75, ax=ax,
        #    edgecolor="gray", palette=d.get('cmap', None))
        #    color='b')

        # .. note: We need to use scatter plot if we want to
        #          assign colors to the markers according to
        #          their value.

        # Using scatter plot
        sns.scatterplot(data=data, x=c,  y='index', s=100,
                        ax=ax, linewidth=0.75, edgecolor='gray',
                        c=data[c], cmap=d.get('cmap', None),
                        norm=d.get('norm', None))

        # Plot vertical lines
        for e in d.get('vline', []):
            vlinebgplot(ax, top=data.shape[0], **e)

        # Configure axes
        ax.set(title=d.get('title', c),
               xlim=d.get('xlim', None),
               xticks=d.get('xticks', []),
               xlabel='', ylabel='')
        ax.tick_params(axis='y', which='both', length=0)
        ax.xaxis.grid(False)
        ax.yaxis.grid(visible=True, which='major',
                      color='gray', linestyle='-', linewidth=0.35)

    # Despine
    sns.despine(left=True, bottom=True)

    # Adjust layout
    plt.tight_layout()
    plt.show()



##############################################################################
# Author: Bernard Hernandez
# Filename: 03-main-create-sari-idxs.py
# Description : This file contains differnent statistics used in time-series.
#               What it mainly does is to format the output of tests provided
#               by external libraries and return them in a dataframe.
#
# TODO: Move it to a module.
#
###############################################################################
# Forces decimals on divisions.
from __future__ import division 

# Libraries
import sys
import numpy as np
import pandas as pd
import statsmodels.api as sm

class SigmoidA:

  def __init__(self, r=200, g=0.5, offset=0.0, scale=1.0,
      percentiles=None, thresholds=None):
    """This function initialises the parameters.

    Parameters
    ----------
    r           : affects the ....
    g           : affects the ....
    offset      : to offset the final sigmoid.
    scale       : to scale the [0,1] sigmoid to the new range.
    percentiles :
    thresholds  :
    """
    self.r = r
    self.g = g
    self.offset = offset
    self.scale = scale
    self.percentiles = percentiles
    self.thresholds = thresholds

  # ---------------------------------------------------------------------------
  #                            HELPER METHODS
  # ---------------------------------------------------------------------------
  def _identifier(self, short=True):
    """This methods describes de model."""
    # Return short description.
    if short:
      per = 'N' if self.percentiles is None else self.percentiles
      ths = 'N' if self.thresholds is None else self.thresholds
      return "Sig(%s, %s)" % (per, ths)
    # Return full description.
    return "%s(r=%s, g=%s, offset=%s, sc=%s, per=%s, ths=%s)" % \
      (self.__class__.__name__, self.r, self.g, self.offset, self.scale, 
        self.percentiles, self.thresholds)

  # ---------------------------------------------------------------------------
  #                                SIGMOID
  # ---------------------------------------------------------------------------
  def approximated_sigmoid(self, x, x_curves=None):
    """This function computes the approximated sigmoid.

    Note: The approximated sigmoid is defined within the interval [0,1].

    .. note: Produces a RuntimeWarning: invalid value encountered in
             true_divide z = (x-cmin) / (cmax-cmin) when the parameter
             x has always the same value.

    Parameters
    ----------
    x: numpy.array
      The values to be converted to weights.
    x_curves: tuple
      The values indicating where the low/upper curves should start.

    Returns
    -------
    r : numpy.array
      The weights.
    """
    # Where the two curves should start.
    cmin, cmax = x_curves if x_curves is not None else [np.min(x), np.max(x)]
    # Sigmoid variables.
    z = (x-cmin) / (cmax-cmin)
    # Approximated sigmoid.
    r = (lambda t: (1+self.r**(-t+self.g))**(-1)) (z)
    # Normalize so it goes between zero and one.
    # This happens when cmin=np.min(x) and cmax=n.max(x)
    r = (r-np.min(r)) / (np.max(r)-np.min(r))
    # Return
    return r

  # ---------------------------------------------------------------------------
  #                            HELPER METHODS
  # ---------------------------------------------------------------------------
  def threshold(self, r, x, thresholds=(None,None)):
    """This function thresholds the r given the values of x.

    Parameters
    ----------
    x: numpy.array
      The values to be converted to weights.
    r: numpy.array
      The weights to be thresholded.
    threshold_low: number;
      The values in x lower will have minimum weight.
    treshold_high: number;
      The values in x higher will have maximum weight.

    Returns
    -------
    """
    # Get thresholds
    threshold_low, threshold_high = thresholds

    # Lower threshold.
    if threshold_low is not None: 
      r = np.where(x<threshold_low, 0, r)

    # Upper threshold.
    if threshold_high is not None: 
      r = np.where(x>threshold_high, np.max(r), r)

    # Return
    return r


  # ---------------------------------------------------------------------------
  #                                WEIGHTS
  # ---------------------------------------------------------------------------
  def weights_standard(self, x):
    """This function returns the weights."""
    return self.approximated_sigmoid(x)


  def weights_percentile(self, x, percentiles):
    """This function returns the weights.

    Parameters
    ----------
    x             :
    x_percentiles :

    Returns
    -------
    weights :
    """
    # Get percentiles
    percentile_low, percentile_high = percentiles
    # Create sigmoid with curves in such percentiles.
    r = self.approximated_sigmoid(x, [np.percentile(x, percentile_low), \
                                      np.percentile(x, percentile_high)])
    # Return
    return r


  def weights(self, x):
    """This cuntion computes the weights.

    Parameters
    ----------
    x :
    x_percentiles    :
    x_threshold_low  :
    x_threshold_high :

    Returns
    -------
    """
    # Get the standard sigmoid or a perso
    if self.percentiles is None:
      r = self.weights_standard(x)
    else :
      r = self.weights_percentile(x, self.percentiles)

    # Treshold the result.
    if self.thresholds is not None:
      r = self.threshold(r, x, self.thresholds)

    # Return
    return self.offset + self.scale*r







if __name__ == '__main__': # pragma: no cover

  # Libraries
  import matplotlib as mpl
  import matplotlib.pyplot as plt
  import statsmodels.robust.norms as norms


  # ------------
  # Basic sample
  # ------------
  # Create SigmoidA instance
  W = SigmoidA(r=200, g=0.5, offset=0.0, scale=1.0)

  # Compute weights
  w = np.ones((1, 5)) * 3
  r = W.weights(w)

  # Show
  print("\nWeights:")
  print(w)
  print(r)

  # ----------------
  # Various samples
  # ----------------

  # Matplotlib options
  mpl.rc('legend', fontsize=6)
  mpl.rc('xtick', labelsize=6)
  mpl.rc('ytick', labelsize=6)
  mpl.rc('font', size=7)

  # Set pandas configuration.
  pd.set_option('display.max_colwidth', 14)
  pd.set_option('display.width', 150)
  pd.set_option('display.precision', 4)

  # Constants
  length = 100

  # Create time-series.
  x = np.linspace(0,100,100)
  f = np.concatenate((np.random.rand(35)*50+50,
                      np.random.rand(30)*50+100,
                      np.random.rand(35)*50+150))

  # Weight Functions
  w_functions = [
    SigmoidA(r=200, g=0.5, offset=0.0, scale=1.0),
    SigmoidA(r=200, g=0.5, offset=0.0, scale=1.0, percentiles=[10,90]),
    SigmoidA(r=200, g=0.5, offset=0.0, scale=1.0, percentiles=[25,75]),
    SigmoidA(r=200, g=0.5, offset=0.0, scale=1.0, percentiles=[40,50]),
    SigmoidA(r=200, g=0.5, offset=0.0, scale=1.0, thresholds=[15,85])]

  # Create figure
  fig, axes = plt.subplots(1,3, figsize=(10,4))

  # Plot frequencies
  axes[1].bar(x, f, color='gray', alpha=0.7, label='Frequency')

  # Plot weights
  for i,W in enumerate(w_functions):
    axes[0].plot(x, W.weights(x), label=W._identifier(short=True))
    axes[2].plot(x, W.weights(f), marker='o', alpha=0.5, 
      markeredgecolor='k', markeredgewidth=0.5, markersize=4, 
      linewidth=0.00, label=W._identifier(short=True))

  # Titles
  axes[0].set_title('Sigmoid Function')
  axes[1].set_title('Frequencies')
  axes[2].set_title('Weights')

  # Legends
  axes[0].legend()
  axes[1].legend()
  axes[2].legend()

  # Tight layout
  plt.tight_layout()

  # Show
  # plt.show()


# Libraries
import numpy as np
import pandas as pd


def create_combinations_v1(d, groupby,
                              col_spe='SPECIMEN',
                              col_lab='LAB_NUMBER',
                              col_org='MICROORGANISM',
                              col_abx='ANTIMICROBIAL',
                              col_sns='SENSITIVITY'):
    """Creates the DataFrame with all combinations.

    .. note:: There might be an issue if there are two different outcomes
              for the same record. For example, a susceptibility test
              record for penicillin (APEN) with R and another one with
              S. Warn of this issue if it appears!

    .. note:: If the data is right and the laboratory numbers are unique per
              isolate then the date is not necessary. However, what if we want
              to keep it? Groupby should at least contain: specimen, microorganism
              and lab_id

    .. note:: How to add all data in addition to the columns manually.

    Parameters
    ----------

    Returns
    --------
    """
    # Libraries
    from itertools import combinations

    # Initialize
    c = []

    # Loop
    for i, g in d.groupby(groupby):
        for x, y in combinations(g.sort_values(by=col_abx).index, 2):
            aux = dict(zip(groupby, i))

            aux.update({
                '%s_x' % col_abx: g.loc[x, col_abx],
                '%s_y' % col_abx: g.loc[y, col_abx],
                '%s_x' % col_sns: g.loc[x, col_sns],
                '%s_y' % col_sns: g.loc[y, col_sns]
            })
            c.append(aux)

    # Create DataFrame
    c = pd.DataFrame(c)

    # Add class
    c['class'] = c['%s_x' % col_sns] + \
                 c['%s_y' % col_sns]
    # Return
    return c






def mutual_info_matrix_v3(x=None, y=None, ct=None):
    """Compute the component information score.

    .. note: Might be inefficient but good for testing.

    .. note: In order to be able to compute the mutual
             information score it is necessary to have
             variation within the variable. Thus, if
             there is only one class, should we return
             a result or a warning?

    Parameters
    ----------
    x: list
        List with the classes
    y: list
        List with the classes

    Returns
    -------
    """
    # Libraries
    from scipy.stats.contingency import crosstab

    def _check_nparray(obj, param_name):
        if obj is not None:
            if isinstance(obj, np.ndarray):
                return obj
            elif isinstance(obj, pd.Series):
                return obj.to_numpy()
            elif isinstance(obj, pd.DataFrame):
                return obj.to_numpy()
            elif isinstance(obj, list):
                return np.array(obj)
            else:
                raise ValueError("""
                       The input parameter '{0}' is of type '{1} which is 
                       not supported. Please ensure it is a np.ndarray."""
                                 .format(param_name, type(obj)))

    # Ensure they are all np arrays
    x = _check_nparray(x, 'x')
    y = _check_nparray(y, 'y')
    ct = _check_nparray(ct, 'ct')

    # Compute contingency
    if ct is None:
        c = crosstab(x,y)
        if isinstance(c, tuple):
            ct = c[-1]   # older scipy
        else:
            ct = c.count # newer scipy

    # Variables
    n = ct.sum()
    pi = np.ravel(ct.sum(axis=1)) / n
    pj = np.ravel(ct.sum(axis=0)) / n

    # Create empty matrix
    m = np.empty(ct.shape)
    m[:] = np.nan

    # Fill with component information score
    with np.errstate(all='ignore'):
        for i in range(m.shape[0]):
            for j in range(m.shape[1]):
                pxy = ct[i,j] / n
                m[i,j] = pxy * np.log(pxy / (pi[i] * pj[j]))

    # Fill with na (lim x->0 => 0)
    m[np.isnan(m)] = 0

    # Return
    return m


def collateral_resistance_index(m):
    """Collateral Resistance Index

    The collateral resistance index is based on the mutual
    information matrix. This implementation assumes there
    are only two classes resistant (R) and sensitive (S).

    .. warning:: Only works for a 2x2 contingency matrix

    Parameters
    ----------
    m: np.array
        A numpy array with the mutual information matrix.
        Also called the contingency matrix.

    Returns
    -------
    """
    return (m[0, 0] + m[1, 1]) - (m[0, 1] + m[1, 0])


def CRI(x, func_mis=mutual_info_matrix_v3):
    """Collateral resistance index

    Parameters
    ----------
    x: pd.Series
        Contains the combinations classes (e.g. SS, SR, RS, RR)

    func_mis: function
        The function to use to compute the contingency matrix from the
        mutual information score. By default it uses the function
        mutual_info_matrix_v3.
    """
    ct = np.array([[x.SS, x.SR], [x.RS, x.RR]])
    m = func_mis(ct=ct)
    return collateral_resistance_index(m)



class ACSI:
    """
    Antimicrobial Collateral Sensitivity Index

    How should we call this index?
       a) Antimicrobial Disjoint Resistance Index
       b) Antimicrobial Collateral Resistance Index
       c) Antimicrobial Collateral Sensitivity Index
       d) Antimicrobial Collateral Resistance Index

    """

    # Attributes
    c_spe = 'SPECIMEN'
    c_org = 'MICROORGANISM'
    c_abx = 'ANTIMICROBIAL'
    c_sns = 'SENSITIVITY'
    c_lab = 'LAB_NUMBER'
    c_dat = 'DATE'

    def __init__(self,
                 column_specimen=c_spe,
                 column_microorganism=c_org,
                 column_antimicrobial=c_abx,
                 column_sensitivity=c_sns,
                 column_laboratory=c_lab):
        """The constructor.

        Parameters
        ----------
        column_specimen: string
            The column name with the specimen

        column_antimicrobial: string
            The column name with the antimicrobial

        column_microorganism: string
            The column name with the microorganism

        column_sensitivity: string
            The column name with the sensitivity

        column_laboratory: string
            The column name with the laboratory id

        Returns
        -------
        none
        """

        # Create dictionary to rename columns
        self.rename_columns = {column_specimen: self.c_spe,
                               column_antimicrobial: self.c_abx,
                               column_microorganism: self.c_org,
                               column_sensitivity: self.c_sns,
                               column_laboratory: self.c_lab}


    def combinations(self, dataframe, **kwargs):
        """Creates the combinations.

        .. note:: In theory the combinations only need to be grouped
                  by laboratory number, however, in order to maintain
                  the rest of the information (like date) we need to
                  pass all of them.
        """
        return create_combinations_v1(dataframe, **kwargs)


    def compute_from_contingency(self):
        pass

    def compute_from_combinations(self):
        pass

    def compute(self, dataframe,
                flag_combinations=False,
                flag_contingency=False,
                groupby=None,
                func_mis=None,
                return_combinations=False):
        """Computes the antimicrobial collateral sensitivity index.

        .. note:: Enable to pass a combinations dataframe.
        .. note:: The lab number should be used only to compute
                  the combinations. It should not be used when
                  computing the ACSI.

        Parameters
        ----------
        dataframe: pd.DataFrame
            A dataframe with the susceptibility test interpretations as columns.

        combinations: boolean
            Indicates whether the variable DataFrame contains susceptibility
            test records (combinations=False) or the antimicrobial combinations
            and the class (combination=True)

        func_mis: function
            The function to use to compute the contingency matrix from the
            mutual information score. By default it uses the function
            mutual_info_matrix_v3.

        Returns
        -------
        pd.Series or pd.DataFrame

        """

        # Do checks

        # Set default groupby
        if groupby is None:
            groupby = [
                self.c_dat,
                self.c_spe,
                self.c_org,
            ]

        # Set default function
        if func_mis is None:
            func_mis = mutual_info_matrix_v3

        # Rename columns
        aux = dataframe.copy(deep=True) \
            .rename(columns=self.rename_columns)

        # Create combinations
        if flag_combinations:
            combinations = aux
        else:
            combinations = self.combinations(aux,
                groupby=groupby + [self.c_lab])

        # Create contingency DataFrame
        contingency = combinations.groupby(
            by=groupby + [
                self.c_abx + '_x',
                self.c_abx + '_y',
                'class'])\
            .size().unstack()

        # Ensure that variables in CRI needed exist
        for s in ['SS', 'RS', 'SR', 'RR']:
            if not s in contingency:
                contingency[s] = 0

        # Compute CRI
        contingency['acsi'] = contingency.fillna(0) \
            .apply(CRI, args=(func_mis,), axis=1)

        # Return
        if return_combinations:
            return contingency, combinations
        return contingency





if __name__ == '__main__':

    # Libraries
    import pandas as pd

    # ----------------------------------
    # Create data
    # ----------------------------------
    # Define susceptibility test records
    susceptibility_records = [
        ['2021-01-01', 'LAB_1', 'BLDCUL', 'ECOL', 'AAUG', 'sensitive'],
        ['2021-01-01', 'LAB_1', 'BLDCUL', 'ECOL', 'ACIP', 'sensitive'],

        ['2021-01-01', 'LAB_2', 'BLDCUL', 'ECOL', 'AAUG', 'sensitive'],
        ['2021-01-01', 'LAB_2', 'BLDCUL', 'ECOL', 'ACIP', 'resistant'],

        ['2021-01-01', 'LAB_3', 'BLDCUL', 'ECOL', 'AAUG', 'sensitive'],
        ['2021-01-01', 'LAB_3', 'BLDCUL', 'ECOL', 'ACIP', 'resistant'],

        ['2021-01-01', 'LAB_4', 'BLDCUL', 'ECOL', 'AAUG', 'resistant'],
        ['2021-01-01', 'LAB_4', 'BLDCUL', 'ECOL', 'ACIP', 'resistant'],

        ['2021-01-02', 'LAB_5', 'BLDCUL', 'ECOL', 'AAUG', 'sensitive'],
        ['2021-01-02', 'LAB_5', 'BLDCUL', 'ECOL', 'ACIP', 'sensitive'],

        ['2021-01-02', 'LAB_6', 'BLDCUL', 'ECOL', 'AAUG', 'sensitive'],
        ['2021-01-02', 'LAB_6', 'BLDCUL', 'ECOL', 'ACIP', 'resistant'],

        ['2021-01-02', 'LAB_7', 'BLDCUL', 'ECOL', 'AAUG', 'resistant'],
        ['2021-01-02', 'LAB_7', 'BLDCUL', 'ECOL', 'ACIP', 'resistant'],

        ['2021-01-03', 'LAB_8', 'BLDCUL', 'ECOL', 'AAUG', 'sensitive'],
        ['2021-01-03', 'LAB_8', 'BLDCUL', 'ECOL', 'ACIP', 'intermediate'],

        ['2021-01-03', 'LAB_9', 'BLDCUL', 'ECOL', 'AAUG', 'resistant'],
        ['2021-01-03', 'LAB_9', 'BLDCUL', 'ECOL', 'ACIP', 'resistant'],
        ['2021-01-03', 'LAB_9', 'BLDCUL', 'SAUR', 'ACIP', 'resistant'],
        ['2021-01-03', 'LAB_9', 'BLDCUL', 'SAUR', 'ACIP', 'resistant'],

        ['2021-01-04', 'LAB_10', 'URICUL', 'ECOL', 'AAUG', 'resistant'],
        ['2021-01-04', 'LAB_10', 'URICUL', 'ECOL', 'ACIP', 'sensitive'],
        ['2021-01-04', 'LAB_10', 'URICUL', 'SAUR', 'AAUG', 'resistant'],
        ['2021-01-04', 'LAB_10', 'URICUL', 'SAUR', 'APEN', 'resistant'],
    ]

    # Create DataFrames
    susceptibility = pd.DataFrame(susceptibility_records,
        columns=['DATE',
                 'LAB_NUMBER',
                 'SPECIMEN',
                 'MICROORGANISM',
                 'ANTIMICROBIAL',
                 'SENSITIVITY'])

    # Format DataFrame
    susceptibility.SENSITIVITY = \
        susceptibility.SENSITIVITY.replace({
            'resistant': 'R',
            'intermediate': 'I',
            'sensitive': 'S'
    })

    # Show
    print("\nSusceptibility:")
    print(susceptibility)

    # .. note: It is important to ensure that there are not susceptibility
    #          test records with contradicting results. For example being
    #          resistant and sensitive at the same time. Integrate this
    #          check in the computation!

    # ---------------------------
    # Create combinations
    # ---------------------------
    # Create combinations
    c = create_combinations_v1(susceptibility,
        groupby=[
            'DATE',
            'LAB_NUMBER',
            'SPECIMEN',
            'MICROORGANISM'
        ])

    print("\nCombinations:")
    print(c)

    # Build contingency
    r = c.groupby([
        'DATE',
        'SPECIMEN',
        'MICROORGANISM',
        'ANTIMICROBIAL_x',
        'ANTIMICROBIAL_y',
        'class']).size().unstack()

    print("Contingency:")
    print(r)

    # Compute CRI
    r['MIS'] = r.fillna(0) \
        .apply(CRI, args=(mutual_info_matrix_v3,), axis=1)

    # Show
    print("\nResult")
    print(r)


    # ------------------------------------------
    # Computes ACSI using class
    # ------------------------------------------

    def show(combinations, contingency, title=None, n=100):
        """Helper function to display outcomes."""
        # Variables
        n_comb = combinations.shape[0]
        n_cont = np.nansum(contingency.to_numpy()[:, :-1])

        if title is None:
            title = 'Grouped By: %s' % str(contingency.index.names[:-2])

        # Display
        print("\n" + "="*n + '\n%s\n'%title + "="*n)
        print("Total combinations: %s" % int(n_comb))
        print("Total contingency:  %s" % int(n_cont))
        print("\nCombinations:")
        print(combinations)
        print("\nContingency:")
        print(contingency)


    # Create ACSI instance
    acsi = ACSI()

    # ---------------
    # Compute overall
    # ---------------
    # .. note:: Why removing LAB_NUMBER returns only
    #           the first letter...?
    # Compute index
    contingency, combinations = \
        acsi.compute(susceptibility,
                     groupby=[],
                     return_combinations=True)

    # Show
    show(combinations, contingency, title='Overall')


    # ---------------
    # Compute by
    # ---------------
    # Compute index
    contingency, combinations = \
        acsi.compute(susceptibility,
                     groupby=['DATE'],
                     return_combinations=True)

    # Show
    show(combinations, contingency, title='By <DATE>')


    # ----------------
    # Compute by pairs
    # ----------------
    # Compute index
    contingency, combinations = \
        acsi.compute(susceptibility,
            groupby=[
                'SPECIMEN',
                'MICROORGANISM'
            ],
            return_combinations=True)

    # Show
    show(combinations, contingency, title='By <SPECIMEN, MICROORGANISM>')

    # -------------------------
    # Compute by date and pairs
    # -------------------------
    # .. note:: It seems that it is important to include all the
    #           parameters when computing the combinations. Otherwise
    #           it might create ill defined combinations. Think this
    #           through...

    # Compute index
    contingency, combinations = \
        acsi.compute(susceptibility,
            groupby=[
                 'DATE',
                 'SPECIMEN',
                 'MICROORGANISM'
            ],
            return_combinations=True)

    # Show
    show(combinations, contingency, title=None)

    # Compute contingency reusing combinations.
    contingency = acsi.compute(combinations.reset_index(),
        groupby=['SPECIMEN'],
        flag_combinations=True,
        return_combinations=False)
    show(combinations, contingency, title=None)

    # Compute contingency reusing combinations.
    contingency = acsi.compute(combinations.reset_index(),
        groupby=['MICROORGANISM'],
        flag_combinations=True,
        return_combinations=False)
    show(combinations, contingency, title=None)

    # -------------------------------------------------------------------------
    # Testing
    # -------------------------------------------------------------------------

    # ---------------------------------------------------------------------
    # Success
    # ---------------------------------------------------------------------
    # .. note: All this examples should succeed. At the moment the code
    #          breaks if gram is not included. This is because the data
    #          we have created has duplicated values for each gram.
    #          Should we consider this within the ASAI?

    # ---------------------------------------------------------------------
    # Errors
    # ---------------------------------------------------------------------
    # .. note: In the examples below, the method acsi is meant to raise
    #          an error either because any of the required missing columns
    #          is missing or because the configuration is not correct.
    print("\n\nHandling errors:")

    # ---------------------------------------------------------------------
    # Warnings
    # ---------------------------------------------------------------------
    # .. note: In the examples below, the method acsi is meant to show a
    #          warning message either no threshold has been specified or
    #          because thresholds have been specified twice.
    print("\n\nShow warnings:")


# Division
from __future__ import division

# Generic libraries
import os
import sys
import glob
import warnings
import numpy as np
import pandas as pd
# import cPickle as pickle # no needed in python 3.x

# Import specific
from os.path import dirname
from pathlib import Path

# Import own module
sys.path.append('../../')

# Import libraries
#import pyamr.utils.io.read as pd_read

# --------------------------------------
# DEFINITION OF DATABASE PATHS
# --------------------------------------
# This paths should be relative to the folder datasets in which they
# are contained. Otherwise it will not work.

# Create dirname
dirname = dirname(__file__)

# ---------
# nhs
# ---------
# Antibiotics
epicimpoc_antibiotics = './nhs/antibiotics/antibiotics.csv'

# Organisms
epicimpoc_organisms = './nhs/organisms/organisms.csv'

# Profiles
epicimpoc_susceptibility_comp = './nhs/susceptibility/complete'
epicimpoc_susceptibility_year = './nhs/susceptibility/by_year'
epicimpoc_susceptibility_type = './nhs/susceptibility/by_cultures'

# Microbiology data


# Other
other_shampoo_sales = './other/shampoo_sales.csv'




# -----------------------------------------------------------------------------
#                              HELPER METHODS
# -----------------------------------------------------------------------------
def make_timeseries():
    """This method creates a hard-coded time series.

    Returns
    -------
    x, y, f:
      The x values, the y values and the frequencies.

    """
    # Create exogenous variable
    x = np.arange(100)
    # Create endogenous variable
    y = np.concatenate((np.arange(50) * 10 + np.random.randn(50) * 20 + 40,
                        np.arange(50) * 2 + np.random.randn(50) * 20 + 400))
    # Create frequency variable
    f = np.concatenate((np.random.rand(35) * 50 + 50,
                        np.random.rand(30) * 50 + 100,
                        np.random.rand(35) * 50 + 150))
    # Return
    return x, y, f


def make_susceptibility():
    """This method returns sample data (Anonymised)"""
    return fixture(name='./nhs/nhs-susceptibility-2009-anonymised.csv')

def load_registry_microorganisms():
    """This method returns the microorganisms registry"""
    return pd.read_csv(Path(dirname) / 'registry' / 'registry_microorganisms.csv')

def load_registry_antimicrobials():
    """This method returns the antimicrobials registry"""
    return pd.read_csv(Path(dirname) / 'registry' / 'registry_antimicrobials.csv')


def load_microbiology_folder(path, folder,
        glob_pattern='susceptibility-*.csv', **kwargs):
    """This method loads the susceptibility data.

    .. note:: It assumes all the susceptibility data is stored in csv
              files whose files name starts with 'susceptibility'. In
              addition, it assumes that the additional iformation is
              is available in files named 'antimicrobials.csv' and
              'microorganisms.csv'

    Parameters
    ----------
    path: string
        The path where the folder is located.
    folder: string
        Name of the folder with the data.
    kwargs:
        Arguments to pass to pd.read_csv

    Returns
    -------
    susceptibility
        The susceptibility test data
    db_abxs
        The registries with the antimicrobials
    db_orgs
        The registry with the microorganisms
    """
    # Define paths
    path = Path("{0}/{1}".format(dirname, path))
    path_sus = path / folder
    path_abx = path / folder / 'antimicrobials.csv'
    path_org = path / folder / 'microorganisms.csv'

    # Load data
    data = pd.concat([ \
        pd.read_csv(f, parse_dates=['date_received'], **kwargs)
            for f in glob.glob(str(path_sus / glob_pattern))])

    # Load databases (registries)
    db_abxs = pd.read_csv(path_abx)
    db_orgs = pd.read_csv(path_org)

    # Return
    return data, db_abxs, db_orgs


def load_data_nhs(folder='susceptibility-v0.0.2', **kwargs):
    """This method loads the susceptibility data.

    """
    return load_microbiology_folder( \
        path='./microbiology/nhs/aggregated/',
        folder=folder, **kwargs)


def load_data_mimic(folder='susceptibility-v0.0.1', **kwargs):
    """This method loads the susceptibility data.

    """
    return load_microbiology_folder( \
        path='./microbiology/mimic/aggregated/',
        folder=folder, **kwargs)

# --------------------------------------
# METHODS TO LOAD DATABASES
# --------------------------------------
# -----------------
# epic impoc basic
# -----------------
"""
def dataset_epicimpoc_antibiotics(**kwargs):
    return pd.read_csv('%s/%s' % (dirname, epicimpoc_antibiotics), *kwargs)

def dataset_epicimpoc_organisms(**kwargs):
    return pd.read_csv('%s/%s' % (dirname, epicimpoc_organisms), *kwargs)


# -----------------------------------
# epic impoc susceptibility test data
# -----------------------------------
def dataset_epicimpoc_susceptibility(**kwargs):
    return pd_read.read_csv('%s/%s' % \
      (dirname, epicimpoc_susceptibility_comp), **kwargs)

def dataset_epicimpoc_susceptibility_year(year='2014', **kwargs):
    return pd_read.read_csv('%s/%s/%s' % \
      (dirname, epicimpoc_susceptibility_year, str(year)), **kwargs)

def dataset_epicimpoc_susceptibility_culture(cultures=['bldcul']):
    pass



def dataset_shampoo_sales(**kwargs):
    return pd.read_csv('%s/%s' % (dirname, other_shampoo_sales), **kwargs)
"""


def fixture(name, **kwargs):
    """Load fixtures

    Parameters
    ----------
    name: string
        The name of the file within the fixtures folder.

    Returns
    --------
    pd.DataFrame
    """
    # Library
    from pathlib import Path

    # Load data
    path = Path(dirname).parent / 'fixtures' / name
    data = pd.read_csv(path, **kwargs)

    # Format
    if 'DATE' in data:
        data.DATE = pd.to_datetime(data.DATE)

    # Return
    return data






if __name__ == '__main__':

    # Import
    import warnings

    # Suppress warnings
    warnings.simplefilter('ignore')

    # Set numpy options
    np.set_printoptions(threshold=np.nan)

    # -----------------------------------
    # Loading default datasets
    # -----------------------------------
    # Load antibiotics
    #antibiotics = dataset_epicimpoc_antibiotics()

    # Load organisms
    #organisms = dataset_epicimpoc_organisms()

    # Load profiles
    #microbiology = dataset_epicimpoc_susceptibility_year(year=2014)


    # Show information
    #print(antibiotics.head(5))
    #print(organisms.head(5))
    #print(len(microbiology))

    #print(dataset_shampoo_sales())



# Libraries
import glob
import time
import pandas as pd

# Import pyAMR
from pyamr.datasets.registries import acronym_series
from pyamr.datasets.load import load_registry_microorganisms
from pyamr.datasets.load import load_registry_antimicrobials

# ---------------------------------
# Methods
# ---------------------------------
def create_microorganisms_lookup_table(orgs):
    """Creates the look up table for the organisms.

    This method uses the information in the organisms dataframe
    and the information in the default microorganisms registry
    to create a unique lookup table for the data.

    Parameters
    ----------
    orgs: pd.DataFrame
        The DataFrame with the organism genus and organism species
        for which the look up table should be created. The DataFrame
        must contain the following columns:
            microorganism_name
            genus
            species

    Returns
    -------
    pd.DataFrame
        Lookup table DataFrame with the following columns:
            'domain'
            'phylum'
            'class'
            'order'
            'family'
            'genus'
            'species'
            'acronym'
            'exists_in_registry'
            'gram_stain'
            'microorganism_code'
            'microorganism_name
    """
    # Check
    if not 'genus' in orgs:
        print("Missing <genus> column.")
    if not 'species' in orgs:
        print("Missing <species> column.")
    if not 'microorganism_name' in orgs:
        print("Missing <microorganism_name> column.")

    # Read microorganisms registry
    reg = load_registry_microorganisms()

    # --------------
    # Step 1
    # --------------
    # First, merge those rows within gram_stain and taxonomy
    # that have equal genus and species. Note that we are
    # only merging if both values exist and are equal.
    # Merge
    orgs = pd.merge(orgs, reg,
        how='left',
        left_on=['genus', 'species'],
        right_on=['genus', 'species']
    )

    # Those merged exist in registry.
    orgs['exists_in_registry'] = orgs.acronym.notna()

    # --------------
    # Step 2
    # --------------
    # Second, for those values whose taxonomy-related columns
    # are null, use the taxonomy information based only on the
    # genus. This does not overwrite step 1.
    # Taxonomy columns
    ctaxonomy = ['domain', 'phylum', 'class', 'order', 'family']

    # Create aux
    aux = pd.merge(orgs[['genus']],
        reg.drop(columns=['species', 'acronym'])\
           .drop_duplicates() \
           .groupby('genus') \
           .head(1),
        how='left',
        left_on=['genus'],
        right_on=['genus']
    )

    # Update orgs
    orgs.update(aux)

    # -------------------
    # Create new acronyms
    # -------------------
    # .. note: In order to be similar to the ones used in
    #          HH hospital, we can pass the minimum value
    #          as lg=1, ls=4 and length of 4 if only one
    #          word.
    # Columns with missing acronyms
    idxs = orgs.acronym.isna()

    # Fill with new acronyms
    orgs.loc[idxs, 'acronym'] = \
        acronym_series(orgs.loc[idxs, 'microorganism_name'].fillna(''),
            exclude_acronyms=reg.acronym.unique().tolist(),
            unique_acronyms=True, verbose=0,
            kwgs_acronym={'sep': '_'})

    # Columns
    keep = ['domain',
            'phylum',
            'class',
            'order',
            'family',
            'genus',
            'species',
            'acronym',
            'exists_in_registry',
            'gram_stain',
            'microorganism_code',
            'microorganism_name',
            'microorganism_name_original']

    # Filter
    orgs = orgs[[c for c in keep if c in orgs]]

    # Return
    return orgs


def create_antimicrobials_lookup_table(abxs):
    """Creates the look up table for the antimicorbials.

    This method uses the information in the antibiotics dataframe
    and the information in the default antimicrobials registry
    to create a unique lookup table for the data.

    Parameters
    ----------
    abxs: pd.DataFrame
        The DataFrame with ... The DataFrame must contain the following
        columns:


    Returns
    -------
    pd.DataFrame
        Lookup table DataFrame with the following columns:
    """
    # Check
    if not 'microorganism_name' in abxs:
        print("Missing <antimicrobial_name> column.")

    # Read microorganisms registry
    reg = load_registry_antimicrobials()

    # --------------
    # Step 1
    # --------------
    # First, merge those rows with equal names.
    # Merge
    abxs = pd.merge(abxs, reg,
        how='left',
        left_on=['antimicrobial_name'],
        right_on=['name']
    )

    # Those merged exist in registry.
    abxs['exists_in_registry'] = abxs.acronym.notna()

    # -------------------
    # Create new acronyms
    # -------------------
    # .. note: In order to be similar to the ones used in
    #          HH hospital, we can pass the minimum value
    #          as lg=1, ls=4 and length of 4 if only one
    #          word.
    # Columns with missing acronyms
    idxs = abxs.acronym.isna()

    # Fill with new acronyms
    abxs.loc[idxs, 'acronym'] = \
        acronym_series(abxs.loc[idxs, 'antimicrobial_name'].fillna(''),
            exclude_acronyms=reg.acronym.unique().tolist(),
            unique_acronyms=True, verbose=0,
            kwgs_acronym={
                'sep': '_',
                'exceptions': []})

    # Columns
    keep = ['name',
            'category',
            'acronym',
            'exists_in_registry',
            'antimicrobial_code']

    # Filter
    abxs = abxs[[c for c in keep if c in abxs]]

    # Return
    return abxs




if __name__ == '__main__':

    # Import
    import csv
    import yaml
    import time
    import logging
    import logging.config
    import pandas as pd

    # Specific
    from pathlib import Path

    # PyAMR specific methods
    from pyamr.datasets.clean import clean_clwsql008
    from pyamr.datasets.clean import clean_legacy
    from pyamr.datasets.clean import clean_mimic

    # Configure logging
    with open('./logging.yaml', 'rt') as f:
        config = yaml.safe_load(f.read())
        logging.config.dictConfig(config)

    # Get logger
    logger = logging.getLogger('dev')

    # ---------------------------------
    # Constant
    # ---------------------------------
    # Time
    timestr = time.strftime("%Y%m%d-%H%M%S")

    # ---------------------------------
    # Methods
    # ---------------------------------
    def strdf(df):
        return "\n\t{0}\n".format(df.to_string().replace('\n', '\n\t'))


    # ---------------------------------
    # Load data
    # ---------------------------------
    # Define path
    path = './nhs/legacy/'
    path = './nhs/clwsql008/'

    tuples = [
        ('./nhs/legacy', clean_legacy),
        ('./nhs/clwsql008', clean_clwsql008),
        #('./nhs/test', clean_clwsql008),
        #('./nhs/test2', clean_legacy),
        #('./mimic/mimic-iv-v0.4', clean_mimic),
        ('./yujia/raw', clean_clwsql008)
    ]

    # Combined data
    combined = []

    # For each tuple
    for path, f_clean in tuples:
        print("Loading... {0}".format(path))
        # Load data (multiple files)
        data = pd.concat([pd.read_csv(f,
            encoding="ISO-8859-1", engine='c')
                for f in glob.glob(path + "/*.csv")])
        # Clean data
        data = f_clean(data)
        # Combine
        combined.append(data)

    # Merge
    data = pd.concat(combined)

    # Basic formatting
    data = data.drop_duplicates()



    # -------------------
    # Anonymise
    # -------------------
    # Create hos_number to id mapper

    #unique = data.patient_id.unique()
    #pid_map = dict(zip(unique, range(len(unique))))

    # Include categories
    #data.patient_id = data.patient_id.map(pid_map)

    #unique = data.patient_hos_number.unique()
    #pid_map = dict(zip(unique, range(len(unique))))

    # Include categories
    #data.patient_id = data.patient_hos_number.map(pid_map)

    # Show
    #logger.info("\nData:\n{0}".format(strdf(data.head(10))))
    logger.info("\nColumns:\n\t{0}\n".format(data.columns))
    logger.info("\nDTypes:\n{0}".format(strdf(data.dtypes)))
    logger.info("\nNaNs:\n{0}".format(strdf(data.isna().sum(axis=0))))


    # ----------------------------------
    # Create Microorganisms LookUp table
    # ----------------------------------
    # Organism columns
    columns = ['microorganism_code',
               'microorganism_name',
               'microorganism_name_original']

    # Extract organisms information from susceptibility
    orgs = data[columns].copy(deep=True)
    orgs = orgs.drop_duplicates(subset=['microorganism_name'])

    # Create genus and species
    orgs[['genus', 'species']] = \
        orgs.microorganism_name.str.split(expand=True, n=1)

    # Format genus and species
    orgs.genus = orgs.genus.str.title()
    orgs.species = orgs.species.str.lower()

    # Sort
    orgs = orgs.sort_values(by=['genus', 'species'])

    # Create microorganisms database
    orgs = create_microorganisms_lookup_table(orgs)


    # ----------------------------------
    # Create Antimicrobials LookUp table
    # ----------------------------------
    # Organism columns
    columns = ['antimicrobial_code',
               'antimicrobial_name']

    # Extract organisms information from susceptibility
    abxs = data[columns].copy(deep=True)
    abxs = abxs.drop_duplicates(subset=['antimicrobial_name'])

    # Format genus and species
    abxs.antimicrobial_name = abxs.antimicrobial_name.str.capitalize()

    # Sort
    abxs = abxs.sort_values(by=['antimicrobial_name'])

    # Create microorganisms database
    abxs = create_antimicrobials_lookup_table(abxs)

    # --------------------------
    # Logging useful information
    # --------------------------
    # This code logs the unique values and the corresponding
    # count for the columns specified in the array. Note that
    # it handles if itdoes not exist (maybe warn?).
    # Report unique values.
    for c in ['sensitivity_code',
              'sensitivity_name',
              'method_code',
              'method_name']:
        if not c in data:
            continue
        # Get value counts
        aux = data[c].value_counts()
        # Log information
        logger.info("\n{0}:\n{1}".format(c, strdf(aux)))

    # This code logs the duplicated values for the subsets
    # included in the array regarding the MICROORGANISMS.
    # Should we also include names?
    for subset in [['microorganism_code'], ['acronym']]:
        # Get duplicates
        idxs_dup = orgs[['microorganism_name',
                         'microorganism_code',
                         'acronym']] \
            .duplicated(subset=subset, keep=False)
        # Log information
        logger.info("\nDuplicated: {0}\n\n{1}".
            format(subset, strdf(orgs[idxs_dup])))

    # This code logs the duplicated values for the subsets
    # included in the array regarding the ANTIMICROBIALS.
    # Should we also include names?
    for subset in [['name'], ['acronym']]:
        # Get duplicates
        idxs_dup = abxs[['name',
                         'antimicrobial_code',
                         'acronym']] \
            .duplicated(subset=subset, keep=False)
        # Log information
        logger.info("\nDuplicated {0}:\n\n{1}"
            .format(subset, strdf(abxs[idxs_dup])))

    # Report duplicated values (antimicrobials)


    """
    # Create basic information
    sensitivity = data.sensitivity_code.value_counts()
    gram_stain = orgs.gram_stain.value_counts()

    # Create duplicates
    aux = orgs[['microorganism_name',
                'microorganism_code',
                'acronym']]

    # Find duplicates
    idxs_dup_code = orgs.duplicated(subset=['microorganism_code'], keep=False)
    idxs_dup_acrm = orgs.duplicated(subset=['acronym'], keep=False)

    # Basic information
    logger.info("\nSensitivity:\n{0}".format(strdf(sensitivity)))
    logger.info("\nGram stain:\n{0}".format(strdf(gram_stain)))
    logger.info("\nDuplicate codes:\n{0}".format(strdf(aux[idxs_dup_code])))
    logger.info("\nDuplicate acronyms:\n{0}".format(strdf(aux[idxs_dup_acrm])))
    """

    # ----------
    # Filter
    # ----------
    # We have to remove those dates in which the date_received is none
    # because otherwise we cannot group the data by year to store it
    # in different files. In addition, the others are also required to
    # have a meaningful susceptibility test record.
    data = data.dropna(how='any',
        subset=['date_received',
                'specimen_name',
                'microorganism_name',
                'antimicrobial_name',
                'sensitivity_name'])

    # ----------
    # Save
    # ----------
    # Columns
    keep = ['date_received',
            'date_outcome',
            'patient_hos_number',
            'laboratory_number',
            'specimen_code',
            'specimen_name',
            'specimen_description',
            'microorganism_code',
            'microorganism_name',
            'antimicrobial_code',
            'antimicrobial_name',
            'method_code',
            'method_name',
            'sensitivity_code',
            'sensitivity_name',
            'mic',
            'reported']

    # Filter
    data = data[[c for c in keep if c in data]]

    # ----------
    # Save
    # ----------
    # Define path
    path = Path('./%s' % timestr)

    # Create path if it does not exist
    path.mkdir(parents=True, exist_ok=True)

    # Save databases
    abxs.to_csv(path / 'antimicrobials.csv', index=False)
    orgs.to_csv(path / 'microorganisms.csv', index=False)

    # Create grouper
    grouper = pd.Grouper(key='date_received', freq='Y')

    # Save susceptibility grouped
    for n, g in data.groupby(grouper):
        # Create filename
        filename = "susceptibility-%s.csv" % n.strftime('%Y')
        # Save
        g.to_csv(path / filename, index=False,
                date_format='%Y-%m-%d %H:%M:%S', # dont check excels! check csvs!
                quoting=csv.QUOTE_ALL)           # QUOTE_NONNUMERIC

    # Logging
    logger.info('The results have been saved in: %s' % path)


###############################################################################
# Author: Bernard Hernandez
# Filename:
# Date: 
# Description:
#
###############################################################################

# Libraries
import os
import sys
import json
import copy
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib.colors as colors
import matplotlib.patches as mpatches
import matplotlib.gridspec as gridspec

from scipy import interp
from matplotlib.colors import ListedColormap

# Own
from pyamr.utils.plot import MidpointNormalize
from pyamr.utils.plot import vlinebgplot


# ------------------------------------------------------------------
# CONSTANTS
# ------------------------------------------------------------------
# These are some pre-defined configurations to facilitate the display
# of some common statistics (R2, jarque-bera, skew) and amr metrics
# (sari, sart, ...) computed in this library.

_DEFAULT_KWGS = {
    'name':'sari',
    'cmap':'Reds',
    'title':'SARI',
    'xlim':[-0.1, 1.1],
    'xticks':[0, 1],
    'kwargs': {
        's':80,
        'vmin':0,
        'vmax':1
    }
}

# Configuration for each columns
_DEFAULT_SARI = {
    'cmap': 'Reds',
    'title': 'SARI',
    'xlim': [-0.1, 1.1],
    'xticks': [0, 1],
    'norm': colors.Normalize(vmin=0, vmax=1),
}

_DEFAULT_SARI_PCT = {
    'cmap': 'Reds',
    'title': 'SARI (%)',
    'xlim': [-1, 101],
    'xticks': [0, 100],
    'norm': colors.Normalize(vmin=0, vmax=100),
}

_DEFAULT_SARI_LR = {
    'cmap': 'Reds',
    'title': 'SARI-lr',
    'xlim': [-0.1, 1.1],
    'xticks': [0, 1],
    'norm': colors.Normalize(vmin=0, vmax=1),
}

_DEFAULT_SART_M = {
    'cmap': 'RdBu_r',
    'title': 'SART (m)',
    'xlim': [-1.2, 1.2],
    'xticks': [-1, 1],
    'norm': colors.Normalize(vmin=-1, vmax=1),
}

_DEFAULT_SART_Y = {
    'cmap': 'RdBu_r',
    'title': 'SART (y)',
    'xlim': [-1.2, 1.2],
    'xticks': [-1, 1],
    'norm': colors.Normalize(vmin=-1, vmax=1),
}

_DEFAULT_R2 = {
    'cmap': 'YlGn',
    'title': 'R2',
    'xlim': [-0.15, 1.1],
    'xticks': [0, 1],
    'norm': colors.Normalize(vmin=0, vmax=1),
}

_DEFAULT_R2ADJ = {
    'cmap': 'YlGn',
    'title': 'R2 Adj',
    'xlim': [-0.15, 1.1],
    'xticks': [0, 1],
    'norm': colors.Normalize(vmin=0, vmax=1),
}

_DEFAULT_JB = {
    'cmap': 'YlGn',
    'title': 'Prob(JB)',
    'xlim': [-0.1, 1.1],
    'xticks': [0, 1],
    'vline': [{'xv': 0.05, 'bg': -0.1}],
    'norm': colors.Normalize(vmin=0, vmax=1),
}

_DEFAULT_JB_PROB = {
    'cmap': 'YlGn',
    'title': 'Prob(JB)',
    'xlim': [-0.1, 1.1],
    'xticks': [0, 1],
    'vline': [{'xv': 0.05, 'bg': -0.1}],
    'norm': colors.Normalize(vmin=0, vmax=1),
}

_DEFAULT_DW = {
    'cmap': 'YlGn_r',
    'title': 'DW',
    'xlim': [-0.4, 4.4],
    'xticks': [0, 2, 4],
    'vline': [{'xv': 0.8, 'bg': 4.4}],
    'norm': colors.Normalize(vmin=0, vmax=4),
}

_DEFAULT_FREQ = {
    'cmap': 'Blues',
    'title': 'Freq',
    'xticks': [0, 40000, 80000],
    'kwargs': {
        's': 80,
        'vmin': 0
    }
}

_DEFAULT_PEARSON = {
    'cmap': 'Reds',
    'title': 'Pearson',
    'xlim': [-1.1, 1.1],
    'xticks': [-1, 0, 1],
    'vline': [{'xv': -0.5, 'bg': -1.1},
              {'xv': 0.5, 'bg': 1.1}],
    'norm': colors.Normalize(vmin=-1, vmax=1),
}

_DEFAULT_KENDALL = {
    'cmap': 'Reds',
    'title': 'Kendall',
    'xlim': [-1.1, 1.1],
    'xticks': [-1, 0, 1],
    'vline': [{'xv': -0.5, 'bg': -1.1},
              {'xv': 0.5, 'bg': 1.1}],
    'norm': colors.Normalize(vmin=-1, vmax=1),
}

_DEFAULT_SPEARMAN = {
    'cmap': 'Reds',
    'title': 'Spearman',
    'xlim': [-1.1, 1.1],
    'xticks': [-1, 0, 1],
    'vline': [{'xv': -0.5, 'bg': -1.1},
              {'xv': 0.5, 'bg': 1.1}],
    'norm': colors.Normalize(vmin=-1, vmax=1),
}

_DEFAULT_SKEW = {
    'cmap': 'Reds',
    'title': 'Skew',
    'xticks':[0],
    'vline': [{'xv': 0}]
}

_DEFAULT_KURTOSIS = {
    'cmap': 'Reds',
    'title': 'Kurtosis',
    'xticks':[3],
    'vline': [{'xv': 3}]
}

_DEFAULT_OMNIBUS = {
    'cmap': 'Reds',
    'title': 'Omnibus',
    'vline': [{'xv': 0.05, 'bg': 1.1}],
    'norm': colors.Normalize(vmin=-1, vmax=1),
}

_DEFAULT_SLOPE = {
    'name': 'x1_tprob',
    'cmap': 'YlGn_r',
    'title': 'P>|t| m',
    'xlim': [-0.1, 1.1],
    'xticks': [0, 1],
    'vline': [{'xv': 0.05, 'bg': 1.1}],
    'norm': MidpointNormalize(vmin=0, vmax=1, midpoint=0.5),
}

_DEFAULT_COEFFICIENT = {
    'name': 'c_tprob',
    'cmap': 'YlGn_r',
    'title': 'P>|t| n',
    'xlim': [-0.1, 1.1],
    'xticks': [0, 1],
    'vline': [{'xv': 0.05, 'bg': 11.1}],
    'norm': MidpointNormalize(vmin=0, vmax=1, midpoint=0.5),
}

_DEFAULT_PERCENT = {
    'name': 'percent',
    'cmap': 'Blues',
    'title': 'Percent',
    'xlim': [7, 9],
    'xticks': [0, 100],
    'vline': [],
}

# Now we combine all of them together. Note
# that the <key> value corresponds to the name
# of the column it should be applied to.
_DEFAULT_CONFIGURATION = {
    'sari': _DEFAULT_SARI,
    'sari_pct': _DEFAULT_SARI_PCT,
    'sart_m': _DEFAULT_SART_M,
    'sart_y': _DEFAULT_SART_Y,
    'dw': _DEFAULT_DW,
    'r2': _DEFAULT_R2,
    'r2_adj': _DEFAULT_R2ADJ,
    'pearson': _DEFAULT_PEARSON,
    'jb': _DEFAULT_JB,
    'jb_prob': _DEFAULT_JB_PROB,
    'percent': _DEFAULT_PERCENT,
    'ptm': _DEFAULT_SLOPE,
    'ptn': _DEFAULT_COEFFICIENT,
    'omnibus': _DEFAULT_OMNIBUS,
    'kurtosis': _DEFAULT_KURTOSIS,
    'skew': _DEFAULT_SKEW,
    'freq': _DEFAULT_FREQ
}






class TableGraph:
    """"""
    # Attributes
    fontsize = 9

    # ---------------------------------------------------------------------------
    #                              HELPER METHODS
    # ---------------------------------------------------------------------------
    def get_map_config(self, title=None,
                             xlim=None,
                             xsym=False,
                             xticks=None,
                             xline=[],
                             vmin=None,
                             vmax=None,
                             vsym=False,
                             cmap='Blues',
                             midpoint=None,
                             s=80,
                             linewidths=0.75,
                             edgecolor='gray',
                             **kwargs):
        """This function helps creating a configuration map.
        """
        # Create new map.
        return { 'ax_kwargs': {
                   'title':title,
                   'xlim':xlim,
                   'xsym':xsym,
                   'xticks':xticks,
                   'xline':xline,
                 },
                 'cmap_kwargs': {
                   'vmin':vmin,
                   'vmax':vmax,
                   'vsym':vsym,
                   'cmap':cmap,
                   'midpoint':midpoint
                 },
                 'scatter_kwargs': {
                   's':s,
                   'linewidths':linewidths,
                   'edgecolor':edgecolor
                 }
               }

    def _init_axes(self, ncolumns):
        """This method initialises the axes.

        Parameters
        ----------
        ncolumns : number of columns.

        Returns
        -------
        axes : array with the corresponding axes.
        """
        # Create grid spec.
        gs = gridspec.GridSpec(1, ncolumns)
        gs.update(left=0.15,
                  right=0.97,
                  bottom=0.12,
                  top=0.85,
                  wspace=0.15)
        # Create subplots.
        return [plt.subplot(gs[j]) for j in range(ncolumns)]

    def _init_conf(self, ncolumns, configuration={},
                                 column_names=None):
        """This method initialises the configuration parameters.

        Parameters
        ----------
        ncolumns      :
        configuration :
        column_names  :

        Returns
        -------
        """
        # Create array with configuration parameters
        v_configuration = []
        # For each column.
        for i in range(ncolumns):
            # Get default configuration.
            config = self.get_map_config()
            # Configuration labelled with an index.
            if i in configuration.keys():
                config.update(configuration[i])
            # Configuration labelled with a name.
            if column_names[i] in configuration.keys():
                config.update(configuration[column_names[i]])
            # Append
            v_configuration.append(config)
        # Return it.
        return v_configuration


    def _vmin_vmax(self, x=None, vmin=None, vmax=None, vsym=True):
        """This function finds the vmin and vmax.

        Parameters
        ----------
        x    : array with values.
        vmin : manually selected vmin.
        vmax : manually selected vmax.

        Returns
        -------
        vmin : the minimum value.
        vmax : the maximum value.
        """
        # Find maximum and minimum.
        if vmin is None: vmin = min(x)
        if vmax is None: vmax = max(x)
        # Symmetric
        if vsym:
            if abs(vmin)>abs(vmax):
                lim = abs(vmin)
            else:
                lim = abs(vmax)
            vmin, vmax = -lim, lim
        # Return
        return vmin, vmax

    def _colormap_info(self, x, cmap='Blues', vmin=None,
                                            vmax=None,
                                            vsym=None,
                                            midpoint=None):
        """This function computes the colormap.

        Parameters
        ----------
        name     : the name of the colormap.
        vmin     : the minimum value to assign a color.
        vmax     : the maximum value to assign a color.
        vsym     : have same color scale when vmin and vmax are not the same.
        midpoint : the midpoint for a cmap with diverging colors.

        Returns
        -------
        colormap : the desired values to pass to scatter.
        """
        # Find vmin and vmax.
        vmin, vmax = self._vmin_vmax(x, vmin, vmax, vsym)
        # Compute basic colormap.
        if isinstance(cmap, str):
            cmap = ListedColormap(sns.color_palette(cmap, len(x)))
        # Compute midpoint colormap.
        if midpoint is not None:
            norm = MidpointNormalize(midpoint=midpoint,
                                     vmin=vmin,
                                     vmax=vmax)
        else:
            norm = None

        # Return
        return x, vmin, vmax, cmap, norm

    # ---------------------------------------------------------------------------
    #                               AXIS METHODS
    # ---------------------------------------------------------------------------
    def _configure_axis(self, x, ax, title=None,
                                   xlim=None,
                                   xsym=False,
                                   xticks=None,
                                   xline=[]):
        """This method configures the axis given external information.

        Parameters
        ----------
        ax : axes to configure.

        Returns
        -------
        ax : the axis that has been modified.
        """
        # Aux variable
        y = np.arange(len(x))
        # Format axis.
        ax.set_yticks(y)                  # Set y ticks.
        ax.set_ylim([-1, len(y)])         # Y axis limit.
        ax.set_facecolor((1, 1, 1))       # Change the background color.
        ax.yaxis.grid(visible=True, which='major', # Create grid with horizontal lines.
                              color='gray',
                              linestyle='-',
                              linewidth=0.35)
        ax.xaxis.grid(visible=True, which='major', # Create grid with horizontal lines.
                              color='gray',
                              linestyle='--',
                              linewidth=0.35)

        # Hide y-labels.
        for j,e in enumerate(ax.get_yticklabels()): # Hide ylabels.
            plt.setp(e, visible=False)

        # Hide all spines.
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.spines['bottom'].set_visible(False)
        ax.spines['left'].set_visible(False)

        # Basic configuration
        if title is not None:  ax.set_title(title)
        if xticks is not None: ax.set_xticks(xticks)
        if xlim is not None:
            vmin, vmax = self._vmin_vmax(vmin=xlim[0], vmax=xlim[1], vsym=xsym)
            ax.set_xlim([vmin, vmax])

        # Add vertical lines.
        for v in xline:
            self.plot_vertical_line(ax, top=len(y)+1, **v)


    def plot_vertical_line(self, ax, top, xv, bg=None, lw=1.0, cb=None):
        """This function adds a vertical line and background

        Parameters
        ----------
        ax  : the axis to plot the line.
        top : the height of the vertical line.
        xv  : the x value where the line should be placed.
        bg  : the x value where the background should be extended.

        Returns
        -------
        """
        # Plot line.
        ax.plot((xv,xv),(-1,top), color='gray',
                                  linestyle='--',
                                  linewidth=lw,
                                  zorder=0)

        # Plot background.
        if bg is not None:
            if cb is None:
                cb = sns.color_palette("Set2", 10)[1]
            ax.fill_between([xv,bg], [-1,-1], [top,top], zorder=0,
                                                   alpha=0.1,
                                                   color=cb)

    def plot_array(self, x, ax=None, ax_kwargs={},
                                   cmap_kwargs={},
                                   scatter_kwargs={},
                                   **kwargs):
        """this method plots a single column.

        Parameters
        ----------
        x              : values to plot.
        ax             : axis to plot the information.
        ax_kwargs      : axis configuration information.
        cmap_kwargs    : cmap configuration information.
        scatter_kwargs : scatter configuration information.

        Returns
        -------
        axis : the axis in which the series is plotted.
        """
        # Create ax if it is not passed.
        if ax is None:
          f, ax = plt.subplots()
        # Create colormap.
        c, vmin, vmax, cmap, norm = self._colormap_info(x, **cmap_kwargs)
        if norm is not None:
            ax.scatter(x=x, y=np.arange(len(x)), c=c,
                       cmap=cmap, norm=norm, **scatter_kwargs)
        else:
            ax.scatter(x=x, y=np.arange(len(x)), c=c,
                       vmin=vmin, vmax=vmax, cmap=cmap,
                       norm=norm, **scatter_kwargs)
        # Configure axis.
        self._configure_axis(x=x, ax=ax, **ax_kwargs)
        # Return axis
        return ax


    def plot_matrix(self, data, axes=None,
                              ylabels=None,
                              clabels=None,
                              configuration={}):
        """This method plots a matrix.

        Parameters
        ----------
        data          :
        ylabels       :
        configuration :
        """
        # Create figure.
        plt.figure()

        # Set style.
        sns.set(style="whitegrid")

        # Get labels.
        ylabels = range(data.shape[0]) if ylabels is None else ylabels
        clabels = range(data.shape[1]) if clabels is None else clabels

        # Initialise the axes.
        if axes is None:
          axes = self._init_axes(data.shape[1])

        # Initliaise the configuration maps.
        conf = self._init_conf(ncolumns=data.shape[1],
                               configuration=configuration,
                               column_names=clabels)

        # Loop and display each column.
        for i, (x,ax,cf) in enumerate(zip(data.T, axes, conf)):
          ax = self.plot_array(x=x, ax=ax, **cf)

        # Set ylabels.
        axes[0].set_yticklabels(ylabels)

        # Show ylabels.
        ticks = axes[0].get_yticklabels()
        for j,e in enumerate(ticks):
            plt.setp(e, visible=True)

        # Set clabels.
        for ax,label in zip(axes,clabels):
          if ax.get_title()=='':
            ax.set_title(label)

        # Return
        return axes



    def plot_dataframe(self, dataframe, axes=None, configuration={}):
        """This method plots the whole table.

        Parameters
        ----------
        dataframe     :
        configuration :

        Returns
        -------
        """
        # Get ylabels and clabels.
        clabels = dataframe.columns.values
        ylabels = dataframe.index.values

        # Convert as matrix.
        if isinstance(dataframe, pd.DataFrame):
          data = dataframe.to_numpy()

        # Plot matrix
        return self.plot_matrix(data, ylabels=ylabels,
                                      clabels=clabels,
                                      axes=axes,
                                      configuration=configuration)


    def plot(self, data, **kwargs):
        """
        """
        if isinstance(data, pd.DataFrame):
          return self.plot_dataframe(dataframe=data, **kwargs)
        if isinstance(data, pd.Series):
          return self.plot_dataframe(dataframe=data.to_frame(), **kwargs)
        if isinstance(data, np.ndarray):
          return self.plot_matrix(data=data, **kwargs)






if __name__ == '__main__':

  """
  
     Example of axis configuration
     -----------------------------
    { 'ax': None,
      'ax_kwargs': {
        'title':'Example 1',
        'xlim':None,
        'xsym':False,
        'xticks':None,
        'xline':[{'xv':0.05, 'bg':-1.0, 'cb':'y'}],
      },
      'cmap_kwargs': {
        'vmin':None,
        'vmax':None,
        'vsym':False,
        'cmap':'RdBu_r',
        'midpoint':2.5
      },
      'scatter_kwargs': { 
        's':80, 
        'linewidths':0.75,
        'edgecolor':'gray'
      }
    }
  """

  # Create data
  data = np.random.randint(100, size=(20, 3))
  dataframe = pd.DataFrame(data)
  dataframe.columns = ['c0','c1','c2']

  # Show
  print("\nData:")
  print(dataframe)

  # Create column configurations
  c0 = TableGraph().get_map_config(title='Column 0', cmap='Reds')
  c2 = TableGraph().get_map_config(title='Column 2', cmap='RdBu_r',
        midpoint=50, xticks=[20,40,60,80],
        xline=[{'xv':25, 'bg':0}, {'xv':75, 'bg':100, 'cb':'y'}])

  # Create configuration
  config = {}
  config['c0'] = c0
  config[2] = c2

  # Example with a DataFrame
  axes = TableGraph().plot(data=dataframe, configuration=config)

  # Example with numpy array
  axes = TableGraph().plot(data=data, configuration=config)

  # Show.
  #plt.show()











class TableGraph2:

  def check_vmin_vmax(self, x, kwargs):
    """This methos check the vmax and vmin for colors.

    Parameters
    ----------

    Returns
    -------
    """
    # Check if there is min and max.
    ismin = 'vmin' in kwargs
    ismax = 'vmax' in kwargs
    # Add to kwargs
    if ismin and ismax: return kwargs
    if ismin and not ismax: kwargs['vmax'] = x.max() 
    if ismax and not ismin: kwargs['vmin'] = x.min()
    if not ismin and not ismax:
      if abs(x.max())>abs(x.min()): 
        l=abs(x.max())
      else:
        l=abs(x.min())
      kwargs['vmin'] = -l
      kwargs['vmax'] = l
    # Return 
    return kwargs


  def add_vline(self, ax, top, xv, bg=None):
    """This function adds a vertical line and background

    Parameters
    ----------

    Returns
    -------
    """
    # Plot line.
    ax.plot((xv,xv),(-1,top), color='gray', 
                              linestyle='--', 
                              linewidth=1,
                              zorder=0)
 
    # Plot background.
    if bg is not None:
      cb = sns.color_palette("Set2", 10)[1]
      ax.fill_between([xv,bg], [-1,-1], [top,top], zorder=0,  
                                                   alpha=0.1, 
                                                   color=cb)
  



  def display_column(self, x, y, cmap, ax, **kwargs):
    """This method displays a single column.

    Parameters
    ----------

    Returns
    -------
    """
    # Check if there is maximum and minimum.
    kwargs = self.check_vmin_vmax(x, kwargs)

    # Normalize colormap.
    if 'cmap_midpoint' in kwargs:
      kwargs['norm'] = \
       MidpointNormalize(midpoint=kwargs['cmap_midpoint'], 
                         vmin=kwargs['vmin'], 
                         vmax=kwargs['vmax'])
      del kwargs['cmap_midpoint']


    # Plot.
    ax.scatter(x=x, 
               y=y, 
               c=x, 
               cmap=cmap,
               edgecolor='gray',
               **kwargs)

    # Configure axis generic.
    ax.set_ylim([-1, len(y)])
    ax.set_xticks([])
    ax.set_yticks(range(len(y)))

    # Common for all.
    #ax.set_axis_bgcolor((1, 1, 1))
    ax.yaxis.grid(visible=True, which='major',
                          color='gray', 
                          linestyle='-',
                          linewidth=0.35)

    # Hide ylabels.
    ticks = ax.get_yticklabels()
    for j,e in enumerate(ticks):
        plt.setp(e, visible=False)


  def plot(self, df, axes, information):
    """This method plots the whole table.

    Parameters
    ----------

    Returns
    -------
    """
    # Set style.
    sns.set(style="whitegrid")
    
    # Loop and display.
    for ax,info in zip(axes,information):
      # Get values.
      x = df[info['name']]
      y = df.index.values
      c = ListedColormap(sns.color_palette(info['cmap'], len(df)))

      # Display column
      self.display_column(x=x, y=y, cmap=c, ax=ax, **info['kwargs'])

      # Basic config.
      if 'title' in info: 
        ax.set_title(info['title'])
      if 'vline' in info:
        for e in info['vline']:
          self.add_vline(ax, top=len(y)+1, **e)
      if 'xlim' in info:
        ax.set_xlim(info['xlim'])
      if 'xticks' in info:
        ax.set_xticks(info['xticks'])

    # Set ylabels.
    #organisms = df['organismCode']   # np.flip(x, axis=0)
    #antibiotics = df['antibioticCode']
    #labels = ["(%-5s,%5s)"%(o,a) for o,a in zip(organisms, antibiotics)]

    #labels = df['antibioticCode']
    #axes[0].set_yticklabels(labels)

    # Show ylabels.
    ticks = axes[0].get_yticklabels()
    for j,e in enumerate(ticks):
        plt.setp(e, visible=True)

    # Return
    return axes





###############################################################################
# Author: Bernard Hernandez
# Filename:
# Date: 
# Description:
#
###############################################################################
# Float division.
from __future__ import division

# General libraries.
import math
import itertools
import numpy as np
import pandas as pd

from sklearn.metrics import mean_absolute_error
from sklearn.metrics import mean_squared_error
from sklearn.metrics import mean_squared_log_error
from sklearn.metrics import median_absolute_error


def _mean_absolute_error(y_true, y_pred):
    """It computes the mean absolute error.
    
    .. note: This function is fully implemented in scikits.

    Parameters
    ----------
    y_true : array-like
      Real values

    y_pred: array-like
      Predicted values

    Returns
    -------
    float
    """
    # Format arrays.
    y_true = y_true.ravel()
    y_pred = y_pred.ravel()
    # Return
    return np.abs(y_true - y_pred).sum() / y_true.shape[0]


def _mean_absolute_percentage_error(y_true, y_pred):
    """It computes the mean absolute error.

    .. note: This function crashes when the y_true value is zero.

    Parameters
    ----------
    y_true : array-like
      Real values

    y_pred: array-like
      Predicted values

    Returns
    -------
    float
    """
    # Format arrays.
    y_true = y_true.ravel()
    y_pred = y_pred.ravel()
    # Return
    return (100. / y_pred.shape[0]) * np.abs((y_true - y_pred) / y_true).sum()


def _mean_directional_accuracy(y_true, y_pred):
    """It computes the mean directional accuracy.

    Parameters
    ----------
    y_true : array-like
      Real values

    y_pred: array-like
      Predicted values

    Returns
    -------
    float
    """
    # Format arrays.
    y_true = y_true.ravel()
    y_pred = y_pred.ravel()
    # Compute
    s1 = np.sign(np.ediff1d(y_true))
    s2 = np.sign(np.ediff1d(y_pred))
    # Return
    return (s1 == s2).sum() / y_true.shape[0]


def _mean_absolute_scaled_error(y_train, y_test, y_pred):
    """It computes the mean absolute scaled error.

    Parameters
    ----------
    y_train : array-like
      Training values

    y_test: array-like
      Testing values

    Returns
    -------
    float
    """
    # Format arrays.
    y_train = y_train.ravel()
    y_test = y_test.ravel()
    y_pred = y_pred.ravel()
    # Compute
    n = y_train.shape[0]
    d = np.abs(np.diff(y_train)).sum() / (n - 1)
    forecast_error = np.abs(y_test - y_pred)
    # Return
    return forecast_error.mean() / d


if __name__ == '__main__': # pragma: no cover

    # Libraries
    import numpy as np

    # Numpy configuration
    np.set_printoptions(precision=2)

    # ---------------------------
    # create data
    # ---------------------------
    # Create time series
    y_true = np.array([10, 11, 12, 11, 12, 11, 13, 12, 13, 14, 15, 16, 18, 17])
    y_pred = np.array([11, 11, 11, 11, 12, 12, 12, 13, 13, 13, 18, 18, 18, 18])

    # Scores
    s1 = mean_absolute_error(y_true, y_pred)
    s2 = _mean_absolute_error(y_true, y_pred)
    s3 = mean_squared_error(y_true, y_pred)
    s4 = mean_squared_log_error(y_true, y_pred)
    s5 = _mean_absolute_percentage_error(y_true, y_pred)
    s6 = _mean_directional_accuracy(y_true, y_pred)
    s7 = _mean_absolute_scaled_error(y_true[:10], y_true[10:], y_pred[10:])
    s10 = median_absolute_error(y_true, y_pred)

    # Compute scores
    print("\nScores:")
    print("-------")
    print('Mean Absolute Error            : %.3f' % s1)
    print('Mean Absolute Error (_)        : %.3f' % s2)
    print('Mean Squared Error             : %.3f' % s3)
    print('Mean Squared Log Error         : %.3f' % s4)
    print('Mean Absolute Percentage (_)   : %.3f' % s5)
    print('Mean Directional Accuracy (_)  : %.3f' % s6)
    print('Mean Absolute Scaled error (_) : %.3f' % s7)
    print('Median Absolute Error          : %.3f' % s10)


##############################################################################
# Author: Bernard Hernandez
# Filename: 03-main-create-sari-idxs.py
# Description : This file contains differnent statistics used in time-series.
#               What it mainly does is to format the output of tests provided
#               by external libraries and return them in a dataframe.
#
# TODO: Move it to a module.
#
###############################################################################
# Forces decimals on divisions.
from __future__ import division

# Libraries
import sys
import warnings
import itertools
import numpy as np
import pandas as pd
import statsmodels.api as sm

# Libraries.
from operator import attrgetter

from statsmodels.tsa.statespace.sarimax import SARIMAX

# Libraries wrapper.
from pyamr.core.regression.wreg import RegressionWrapper


# --------------------------------------------------------------------------
#                              helper methods
# -------------------------------------------------------------------------- 
def _cast_float(e):
    """Casts an element to float when feasible.
    """
    try:
        return float(e)
    except:
        return e


class SARIMAXWrapper(RegressionWrapper):
    """Description

    """

    # --------------------------------------------------------------------------
    #                              overriden
    # --------------------------------------------------------------------------
    def _identifier(self):
        """This method returns the name to idenfy the model.
        """
        return "%s%sx%s [%s,%s]" % (getattr(self, '_name', None),
                                    getattr(self, 'order', None),
                                    getattr(self, 'seasonal_order', '(None)'),
                                    getattr(self, 'trend', None),
                                    getattr(self, 'exog', None) is not None)

    # --------------------------------------------------------------------------
    #                        parameters from summary
    # --------------------------------------------------------------------------
    def _params_from_summary(self):
        """This method returns a dictionariy with the parameters from summary.
        """
        # Format summary
        summary = self._raw.summary().as_csv()
        summary = summary.split("\n", 1)[1]  # Remove first line.
        summary = summary.replace("\n", ",")  # Replace \n by comma.

        # Split in elements.
        elements = summary.split(",")
        elements = [_cast_float(e.strip()) for e in elements]

        # Create series.
        d = {}

        # Add parameters.
        d['s_jb_value'] = elements[-13]
        d['s_jb_prob'] = elements[-9]
        d['s_skew'] = elements[-5]
        d['s_Q_value'] = elements[-15]
        d['s_Q_prob'] = elements[-11]
        d['s_H_value'] = elements[-7]
        d['s_H_prob'] = elements[-3]
        d['s_kurtosis'] = elements[-1]
        d['s_heteroskedasticity'] = elements[-7]
        d['s_omnibus_value'] = None
        d['s_omnibus_prob'] = None

        # Return
        return d

    def evaluate(self, alpha=0.05):
        """This method creates a dictioanry with the relevant parameters.

        @see: statsmodels.Sarimax
        @see: statsmodels.SarimaxResults

        Parameters
        ----------
        alpha : significance level

        Returns
        -------
        dictionary : dictionary with the parameters.
        """
        # Create series.
        d = {}

        # Add generic metrics.
        d['aic'] = self._raw.aic
        d['bic'] = self._raw.bic
        d['hqic'] = self._raw.hqic
        d['llf'] = self._raw.llf

        # Create params information.
        params_data = zip(self._raw.data.param_names,
                          self._raw.params,
                          self._raw.bse,
                          self._raw.zvalues,
                          self._raw.pvalues,
                          self._raw.conf_int(alpha))

        # Add coefficients statistics to series.
        for name, coef, std, tvalue, pvalue, (cil, ciu) in params_data:
            d['%s_%s' % (name, 'coef')] = coef
            d['%s_%s' % (name, 'std')] = std
            d['%s_%s' % (name, 'tvalue')] = tvalue
            d['%s_%s' % (name, 'tprob')] = pvalue
            d['%s_%s' % (name, 'cil')] = cil
            d['%s_%s' % (name, 'ciu')] = ciu

        # Further statistics.
        d.update(self._params_from_summary())
        d.update(self._resid_stats())

        # Extra useful infomation
        d['converged'] = self._raw.mle_retvals['converged']

        # Return
        return d

    # --------------------------------------------------------------------------
    #                             summary method
    # --------------------------------------------------------------------------
    def as_summary(self, **kwargs):
        """This method creates a summary string.
        """
        # Elements to split by.
        find = "=" * 78

        # Split and fill.
        smry = find.join(self._raw.summary(**kwargs).as_text().split(find)[:-1])
        smry = smry.split("\n")
        smry[-6] = smry[-6].replace('=', '', 5)
        smry[-5:] = [v.replace(' ', '', 5) for v in smry[-5:]]
        smry = "\n".join(smry[:-1])

        # Variables.
        om, omp, dw = 0.0, 0.0, self.m_dw
        jb, jbp = self.m_jb_value, self.m_jb_prob
        nm, nmp = self.m_nm_value, self.m_nm_prob
        skew, kurt = self.m_skew, self.m_kurtosis

        # Add in new lines.
        smry += "\n%s\n%s\n%s\n" % ("=" * 78, "Manual".center(78, ' '), "-" * 78)
        smry += "Omnibus:    %#25.3f   Durbin-Watson: %#23.3f\n" % (om, dw)
        smry += "Normal (N): %#25.3f   Prob(N):       %#23.3f\n" % (nm, nmp)
        smry += "=" * 78 + "\n"
        smry += "Note that JB, P(JB), skew and kurtosis have different values.\n"
        smry += "Note that Prob(Q) tests no correlation of residuals."

        # Return
        return smry

    # ---------------------------------------------------------------------------
    #                           prediction method
    # ---------------------------------------------------------------------------
    def get_prediction(self, start=None, end=None, alpha=0.05, **kwargs):
        """This method calls the get prediction function in ARIMA.

        Parameters
        ----------
        start : int (optional)
          The time t to start the prediction

        end : int (optional)
          The time t to end the prediction

        dynamic : bool
          The dynamic keyword affects in-sample prediction. If dynamic is false
          then the in-sample lagged values are used for prediction. if dynamic
          is true, then in-sample forecasts are used in place of lagged dependent
          variables. The first forecasted value is start.

        alpha : float
          The alpha value when creating the norm point percent function to
          compute the confidence intervals for the forecast predictions.

        typ: string-like, options = {linear, levels}
          Wether to predict the original levels (levels) or predict in terms of
          the differenced endogenous variables.

        Returns
        -------
        matrix :
        """
        # Predict levels by default in differenced timeseries.
        if hasattr(self._raw.model, 'k_diff'):
            # Return original levels
            if not 'typ' in kwargs:
                kwargs['typ'] = 'levels'

        # Compute prediction.
        prediction = self._raw.get_prediction(start=start, end=end, **kwargs)

        # Compute time.
        time = self._time(start=start, end=end)

        # Divide predictions in insample/forecast
        pred_in = prediction.predicted_mean[time < len(self.endog)]

        # Get plotting values.
        mean = prediction.predicted_mean
        cito = prediction.conf_int()  # .as_matrix()

        # Add the insample confidence interval. Note that the prediction
        # conf_int implemented above computes the forecast error. However,
        # there are some observatons (pred_in) which have been seen during
        # model fitting and therefure they are not pure forecasts.
        cito[:len(pred_in), :] = self.conf_int_insample(pred_in, alpha=alpha)

        # Get plotting values.
        return np.column_stack((time, mean, cito)).T

    # --------------------------------------------------------------------------
    #                           auto method
    # --------------------------------------------------------------------------
    def auto(self, endog, exog=None, ic='bic', max_ar=3, max_ma=3, max_d=1,
             max_P=0, max_D=0, max_Q=0, list_s=[12], warn='ignore',
             trends=['n', 'c', 't', 'ct'], return_fits=False, verbose=0,
             converged=True, **kwargs):
        """This method finds the best arima through bruteforce.

        Note: It uses grid search through the base wrapper.

        Parameters
        ----------
        endog : array-like
          The endogenous variable (aka. time series data)

        exog : array-like
          The exogenous variable (by default is the time t starting in 0)

        ic : string
          The information criteria which should be any of the params which
          are set in the wrapper (see method _init_result).

        max_ar : number
          The maximum autorregresive order to inspect.

        max_ma : number
          The maximum moving average order to inspect.

        max_d : number
          The maximum difference to inspect.

        max_P : number

        max_D : number

        max_Q : number

        list_s : array-like

        warn : string, options = {ignore, raise}
          Whether or not to ignore the warnings raised

        return_fits: bool
          Whether or not to return all the fits. If false only the best model
          is returned. If true all the models and the best model are returned
          separately.

        converged: bool
          Whether or not to return only converging models. If false all models
          are returned. If true, only those models that converged are returned.

        Returns
        -------
        """
        # Compute all the orders to perform bruteforce search.
        # Note that some orders are not valid, these will throw an
        # exception that will be catched in the base wrapper
        # grid_search method. For instance (0,0,0) is invalid.
        orders = list(itertools.product(*[np.arange(0, max_ar + 1),
                                          np.arange(0, max_d + 1),
                                          np.arange(0, max_ma + 1)]))

        # Similar than previous guide but with seasonal orders.
        seasonal_orders = list(itertools.product(*[np.arange(0, max_P + 1),
                                                   np.arange(0, max_D + 1),
                                                   np.arange(0, max_Q + 1),
                                                   list_s]))

        # Parameters
        grid_params = {'exog': [exog],
                       'endog': [endog],
                       'order': orders,
                       'seasonal_order': seasonal_orders,
                       'trend': trends,
                       'disp': [0]}

        # Perform grid search
        with warnings.catch_warnings():
            # What to do with the warnings
            warnings.filterwarnings(warn)
            # Perform grid search.
            wrappers = self.grid_search(grid_params=grid_params, verbose=verbose)

        # Keep only those that converged.
        if converged:
            wrappers = [w for w in wrappers if w.converged]

        # Return all
        if return_fits:
            return wrappers, min(wrappers, key=attrgetter(ic))

        # Return best
        return min(wrappers, key=attrgetter(ic))

    # ---------------------------------------------------------------------------
    #                             OTHER METHODS
    # ---------------------------------------------------------------------------
    def save(self, **kwargs):
        """This method saves the information.
        """
        self._raw.save(**kwargs)

    def load(self, **kwargs):
        """This method loads the information.
        """
        # Load.
        self._raw = sm.load(**kwargs)
        self._resid = self._raw.resid
        self._result = self._init_result()
        self._config = self._init_config()
        # Return.
        return self


if __name__ == '__main__': # pragma: no cover

    # Import
    import sys
    import warnings
    import pandas as pd
    import matplotlib as mpl
    import matplotlib.pyplot as plt

    # Set pandas configuration.
    pd.set_option('display.max_colwidth', 14)
    pd.set_option('display.width', 140)
    pd.set_option('display.precision', 4)

    # Import own libraries
    from pyamr.datasets.load import make_timeseries

    # ----------------------------
    # set basic configuration
    # ----------------------------
    # Matplotlib options
    mpl.rc('legend', fontsize=6)
    mpl.rc('xtick', labelsize=6)
    mpl.rc('ytick', labelsize=6)

    # Set pandas configuration.
    pd.set_option('display.max_colwidth', 14)
    pd.set_option('display.width', 150)
    pd.set_option('display.precision', 4)

    # ----------------------------
    # create data
    # ----------------------------
    # Create timeseries data
    x, y, f = make_timeseries()

    # ----------------------------
    # fit the model
    # ----------------------------
    # Create specific sarimax model.
    sarimax = SARIMAXWrapper(SARIMAX).fit(endog=y[:80],
                                          exog=None,
                                          trend='ct',
                                          seasonal_order=(1, 0, 1, 12),
                                          order=(0, 1, 1),
                                          disp=0)

    # Print series.
    print("\nSeries:")
    print(sarimax.as_series())

    # Print summary.
    print("\nSummary:")
    print(sarimax.as_summary())

    # -----------------
    # Save & Load
    # -----------------
    # File location
    # fname = '../../examples/saved/sarimax-sample.pickle'

    # Save
    # sarimax.save(fname=fname)

    # Load
    # sarimax = SARIMAXWrapper().load(fname=fname)

    # -------------
    #  Example I
    # -------------
    # This example shows how to make predictions using the wrapper which has
    # been previously fitted. It also demonstrateds how to plot the resulting
    # data for visualization purposes. It shows two different types of
    # predictions:
    #    - dynamic predictions in which the prediction is done based on the
    #      previously predicted values. Note that for the case of ARIMA(0,1,1)
    #      it returns a line.
    #    - not dynamic in which the prediction is done based on the real
    #      values of the time series, no matter what the prediction was for
    #      those values.

    # Variables.
    s, e = 50, 120

    # Compute predictions
    preds_1 = sarimax.get_prediction(start=s, end=e, dynamic=False)
    preds_2 = sarimax.get_prediction(start=s, end=e, dynamic=True)

    # Create figure
    fig, axes = plt.subplots(1, 2, figsize=(8, 3))

    # Plot non-dynamic
    # ----------------
    # Plot truth values.
    axes[0].plot(y, color='#A6CEE3', alpha=0.5, marker='o',
                 markeredgecolor='k', markeredgewidth=0.5,
                 markersize=5, linewidth=0.75, label='Observed')

    # Plot forecasted values.
    axes[0].plot(preds_1[0, :], preds_1[1, :], color='#FF0000', alpha=1.00,
                 linewidth=2.0, label=sarimax._identifier())

    # Plot the confidence intervals.
    axes[0].fill_between(preds_1[0, :], preds_1[2, :],
                         preds_1[3, :],
                         color='#FF0000',
                         alpha=0.25)

    # Plot dynamic
    # ------------
    # Plot truth values.
    axes[1].plot(y, color='#A6CEE3', alpha=0.5, marker='o',
                 markeredgecolor='k', markeredgewidth=0.5,
                 markersize=5, linewidth=0.75, label='Observed')

    # Plot forecasted values.
    axes[1].plot(preds_2[0, :], preds_2[1, :], color='#FF0000', alpha=1.00,
                 linewidth=2.0, label=sarimax._identifier())

    # Plot the confidence intervals.
    axes[1].fill_between(preds_2[0, :], preds_2[2, :],
                         preds_2[3, :],
                         color='#FF0000',
                         alpha=0.25)

    # Configure axes
    axes[0].set_title("SARIMAX non-dynamic")
    axes[1].set_title("SARIMAX dynamic")

    # Format axes
    axes[0].grid(True, linestyle='--', linewidth=0.25)
    axes[1].grid(True, linestyle='--', linewidth=0.25)

    # Legend
    axes[0].legend()
    axes[1].legend()

    # plt.show()

    # -------------------------------
    # Example II - AUTO
    # -------------------------------
    # This example shows how to use auto to find the best overall model using
    # a particular seletion criteria. It also demonstrates how to plot the
    # resulting data for visualization purposes. Note that it only prints
    # the top best classifier according to the information criteria.
    #
    # To do: Review the dynamic=True.

    sarimax = SARIMAXWrapper(estimator=SARIMAX)

    # Find the best arima model (bruteforce).
    models, best = sarimax.auto(endog=y[:80], ic='bic',
                                max_ar=2,
                                max_ma=3,
                                max_d=1,
                                max_P=1,
                                max_D=0,
                                max_Q=1,
                                list_s=[1],
                                verbose=1,
                                return_fits=True)

    # Sort the list (from lower to upper)
    models.sort(key=lambda x: x.bic, reverse=False)

    # Summary
    summary = sarimax.from_list_dataframe(models)

    # Show summary
    print("\nSummary:")
    print(summary[['sarimax-order',
                   'sarimax-seasonal_order',
                   'sarimax-trend',
                   'sarimax-aic',
                   'sarimax-bic']])

    # Create figure
    fig, axes = plt.subplots(3, 3, figsize=(10, 6))
    axes = axes.flatten()

    # Loop for the selected models
    for i, estimator in enumerate(models[:9]):
        # Show information
        print("%2d. Estimator (bic=%.2f): %s " % \
              (i, estimator.bic, estimator._identifier()))

        # Get the predictions
        preds = estimator.get_prediction(start=s, end=e, dynamic=False)

        # Plot truth values.
        axes[i].plot(y, color='#A6CEE3', alpha=0.5, marker='o',
                     markeredgecolor='k', markeredgewidth=0.5,
                     markersize=5, linewidth=0.75, label='Observed')

        # Plot forecasted values.
        axes[i].plot(preds[0, :], preds[1, :], color='#FF0000', alpha=1.00,
                     linewidth=2.0, label=estimator._identifier())

        # Plot the confidence intervals.
        axes[i].fill_between(preds[0, :], preds[2, :],
                             preds[3, :],
                             color='#FF0000',
                             alpha=0.25)

        # Configure axes
        axes[i].legend(loc=3)
        axes[i].grid(True, linestyle='--', linewidth=0.25)

    # Set superior title
    plt.suptitle("Dynamic predictions for SARIMAX")

    # Show
    # plt.show()



###############################################################################
# Author: Bernard Hernandez
# Filename: organismspy
# Date: 02/09/2015
# Description:
# 
# This script creates a map to solve possible issues of misspelled antibiotics,
# or differences as upper/lower cases, unnecessary spaces, missing codes, ...
# Hence, it creates a conversion between the original antibiotic name to an
# homogenous antibiotic name, and the original code and an homogeneous code.
#
###############################################################################


# Generic libraries.
import sys
import datetime
import numpy as np
import pandas as pd
import re

"""
# Own libraries.
sys.path.append('../../../../modules/')
import settings.organisms as ORG
import settings.antibiotics as ANT
import settings.microbiology as MBL
import others.io.read as pd_read
import others.clean.generic as pd_clean_gen
"""

# IMPORTANT
# =========
# Note that some of the constants used to indicate which columns should be
# lowercase, uppercase, ... are stated in the settings modules which can be
# found in modules/settings/organisms.py. In addition, the rename_map
# stores in modules/settings/microbiology.py should contain the mapping
# between the column names with the organism name and code and the
# standard: organismNameOrig and organismCodeOrig



class OrganismsTable:
  """
  """

  # ----------
  # Constants
  # ----------
  # Columns identifiers for the name and the code. Ensure that there is 
  # a conversion in the file modules/settings/microbiology.py from the
  # input data file column names to 'organismCodeOrig' and
  # 'organismNameOrig'. The different letters indicate:
  # O - the original value that will be stored in the antibiotics table
  # F - the formated value that will be stored in the antibiotics table
  # P - the value of the column in the input data.
  nameO, codeO = 'organismNameOrig', 'organismCodeOrig'
  nameF, codeF = 'organismName', 'organismCode' 
  nameS, codeS = 'specieName', 'specieCode'

  # The columns that should be read from the file. Note that the original
  # files are huge and might not be stored in memory. Therefore reducing
  # the columns kept in memory helps.
  usecols = [nameO, codeO] 

  # Length for automatic codes.
  len_sp = 8
  len_tp = 8

  # Constructor
  def __init__(self):
    txt = "Check that the columns with the ORGANISM name and code have "
    txt+= "mapped in the rename_map variable in the file "
    txt+= "module/settings/microbiology.py\n"
    print(txt)



  #------------------------------------------------------
  #                  PRIVATE METHODS
  #------------------------------------------------------
  def _acronym(self, words):
    """This method return the acronym of several words.
    """
    return "".join(e[0] for e in words)


  def _compute_main(self, name):
    """This method creates the main part of the code.
    """
    words = name.split(" ")
    if 'sp.'in name:    return name[:self.len_sp]
    elif len(words)==1: return name[:self.len_sp]
    elif len(words)==2: return words[0][0]+words[1][:self.len_tp]
    else:               
      if 'beta-haemolytic' in name:
        tp = re.sub('[\W_]+', '', words[-1])
        return self._acronym(words[:-1])+tp
      else:
        return self._acronym(words)


  def _compute_numb(self, name):
    """This method computes the numbers.
    """
    if 'second' in name: return 2
    if 'third' in name: return 3
    if '2nd' in name: return 2
    if '3rd' in name: return 3
    return None


  def _compute_automatic_code(self, row, len_sp=8, len_tp=8):
    """This method computes the automatic code (starting with A_). 
    """
    # Compute.
    main = self._compute_main(row['organismName'])
    numb = self._compute_numb(row['organismNameOrig'])
    if numb is None: code = "A_%s" % main
    else:            code = "A_%s%s" % (main,numb)  
    # Return
    return code.upper() 

  def _compute_organism_codes(self, df):
    """This method computes missing organism codes.
    """
    # compute codes
    for idx,elm in df.iterrows():
      # It already has a code.
      if not pd.isnull(elm[self.codeO]): continue
      # Needs an automatic code but name is empty.
      if pd.isnull(elm[self.nameO]): continue
      if pd.isnull(elm[self.nameF]): continue
      # Compute automatic code
      df.loc[idx, self.codeF] = self._compute_automatic_code(elm)
    # return
    return df

  def _compute_specie_codes(self, df):
    """This method computes missing specie codes.
    """
    # Find species
    species_rows = df['organismName'].str.contains('sp.')
    species_vals = df['organismName'][species_rows].unique()
    for sp in species_vals:
      sp_name = sp.split(" ")[0]
      sp_code = sp_name.upper()
      sp_rows = df['organismName'].str.contains(sp_name)
      df.loc[sp_rows,'specieName'] = sp_name
      df.loc[sp_rows,'specieCode'] = 'A_%s' % sp_code
    return df

  #------------------------------------------------------
  #                  PUBLIC METHODS
  #------------------------------------------------------
  def merge_table(self, df1, df2, on, conflicts):
    """This function merges the two basic columns name and code.

    Parameters
    ----------
    df1 : the old dataframe with name and code.
    df2 : the new dataframe with name and code.

    Returns
    -------
    df_m : merged dataframe.
    df_c : conflicts dataframe.
    """
    # Merged dataframe
    df_m = pd.merge(df1, df2, on=on, how='outer')
    for c1 in df_m.columns[1:]:
      if not c1.endswith('_x') and not c1.endswith('_y'): continue
      c2 = c1[:-1]+'x' if c1.endswith('_y') else c1[:-1]+'y'
      df_m[c1].fillna(df_m[c2], inplace=True)
    # Find conflicts
    c1, c2 = conflicts+"_x", conflicts+"_y"
    df_c = df_m[df_m[c1]!=df_m[c2]]
    df_c = df_c[[on, c1, c2]]
    # Remove duplicated columns (y)
    for c in df_m.columns:
      if c.endswith('y'):
        del df_m[c]
    # Remove column names endings.
    for i,c in enumerate(df_m.columns.values):
      if c.endswith('x'):
        df_m.columns.values[i] = c[:-2]
    # Return
    return df_m, df_c


  def create_table(self, data_path):
    """This method creates the table from input data

    Parameters
    ----------

    Returns
    -------
    """
    """
    # Read data.
    df = pd_read.read_data(ftype='csv', 
                           path=data_path, 
                           rename_map=MBL.rename_map,          # renaming cols
                           keep_cols=self.usecols,             # keep cols
                           std_cols=[self.nameO, self.codeO],  # std cols 
                           low_memory=False)

    # Unique combinations (NaN are not grouped so they are set to -1).
    df_u = df.replace(np.nan, "None")
    df_u = df_u.groupby([self.nameO, self.codeO]).size()
    df_u = df_u.reset_index().rename(columns={0:'count'})
    df_u = df_u.replace("None", np.nan)
    df_u[self.nameF] = df_u[self.nameO]
    df_u[self.codeF] = df_u[self.codeO]
    df_u.dropna(subset=[self.nameO], inplace=True)
    #df_u = df_u.reindex(columns=ANT.database_cols)
    del df_u['count']

    # Cleaning data (order matters).
    df_u = pd_clean_gen.lettercase(df_u, ORG.to_lowercase, 'lower')
    df_u = pd_clean_gen.lettercase(df_u, ORG.to_uppercase, 'upper')
    df_u = pd_clean_gen.delete_spaces(df_u, ORG.dl_spaces)
    df_u = pd_clean_gen.ending_specie(df_u, ORG.specie_ending)
    df_u = pd_clean_gen.specie_abbreviation(df_u, ORG.specie_abbreviation)
    df_u = pd_clean_gen.delete_parenthesis(df_u, ORG.dl_parenthesis)
    df_u = pd_clean_gen.delete_numbers(df_u, ORG.dl_numbers)

    # Computing automatic codes.
    df_u = self._compute_organism_codes(df_u)
    df_u = self._compute_specie_codes(df_u)

    # Drop duplicates keeping the one in which the code is not NAN. For that
    # purpose we sort by values (therefore sending nans to the end) and then
    # use drop_duplicates() removing the last instance found.
    df_u = df_u.sort_values(by=self.codeO)
    df_u = df_u.drop_duplicates(subset=[self.nameO], keep='first')
    df_u = df_u.sort_values(by=self.nameO)

    # Name of columns
    df_u.columns = [self.nameO, self.codeO, 
                    self.nameF, self.codeF,
                    self.nameS, self.codeS]

    # Return
    return df_u
    """
    pass


  def compute(self, input_path, output_path):
    """This method computes/updates a table from data.

    Parameters
    ----------
    input_path  : the path with the data.
    output_path : the path to store the table.

    Returns
    -------

    """
    """
    # Create table using input data.
    tnew = self.create_table(input_path)
    told = pd_read.read_data(ftype='csv', path=output_path) 

    # Merge tables.
    dfm, dfc = tnew, None
    if told is not None:
      cols = [self.nameO, self.codeF]
      dfm, dfc = self.merge_table(told, tnew, self.nameO, self.codeF)
    
    # Sort and save.
    dfm.sort_values(by=self.nameO, inplace=True)
    dfm.to_csv(output_path, index=False)

    # Show information.
    self.show_information(dfm, dfc, output_path)

    # return
    return output_path
    """
    pass



  def show_information(self, df, dfc, pathname):
    """This method shows important information.

    Parameters
    ----------

    Returns
    -------
    """
    # Text displayed as info when merging different files.
    txt_init = "Antibiotics table conversion created!\n"
    txt_merge = "\nThe following conflicts have been found. Please take action "
    txt_merge+= "by opening the automatically created antibiotic csv file "
    txt_merge+= "(antibotic database) and solving such conflicts before "
    txt_merge+= "continuing (by default the old values have been kept):"

    # Variables
    dup_nameO = df[df.duplicated(self.nameO)]
    dup_codeO = df[df.duplicated(self.codeO)]
    dup_codeF = df[df.duplicated(self.codeF)]

    # Print text init
    print("-"*80 + "\n"+txt_init)
    print("Different Original Names: %s" % len(df[self.nameO].unique()))
    print("Different Formated Names: %s " % len(df[self.nameF].unique()))
    print("Different Organism Codes: %s" % len(df[self.codeF].unique()))
    print("Different Species Names: %s" % len(df[self.nameS].unique()))
    print("Different Species Codes: %s\n" % len(df[self.codeS].unique()))

    # Print duplicates
    if len(dup_nameO)>0: print(dup_nameO)
    if len(dup_codeO)>0: print(dup_codeO)
    if len(dup_codeF)>0: print(dup_codeF)

    # Print conflicts
    if dfc is not None:
      if len(dfc)>0:
        print("Conflicts:")
        print(dfc)
        print(txt_merge)

    print("Please revise: %s" % pathname)














if __name__ == '__main__':

  # Import own module
  sys.path.append('../../../')

  # Import specific libraries
  from pyamr.datasets import load

  # Constants.
  goback = "../../../../"
  input_path = goback + "data/raw/microbiology/csv/luke02/"
  output_path = goback + "data/tables/antibiotics.csv"


  # ---------------------
  # load data
  # ---------------------
  # Load data
  data = load.dataset_epicimpoc_susceptibility_year(nrows=1000)

  # ---------------------
  # create builder
  # ---------------------
  # Builter
  #builder = AntibioticsTable()
  #print(data.columns)

  #builder.fit()
  #import sys
  #sys.exit()

  # Constants.
  goback = "../../../../"
  input_path = goback + "data/raw/microbiology/csv/luke02"
  output_path = goback + "data/tables/organisms.csv"

  # Object.
  builder = OrganismsTable()

  # Format.
  builder.compute(input_path, output_path)


##############################################################################
# Author: Bernard Hernandez
# Filename: 03-main-create-sari-idxs.py
# Description : This file contains differnent statistics used in time-series.
#               What it mainly does is to format the output of tests provided
#               by external libraries and return them in a dataframe.
#
# TODO: Move it to a module.
#
###############################################################################
# Forces decimals on divisions.
from __future__ import division

# Libraries
import sys
import copy
import numpy as np
import pandas as pd
import statsmodels.api as sm
from statsmodels.tsa.arima_model import ARIMA
from statsmodels.tsa.statespace.sarimax import SARIMAX

import scipy.stats as stats

# Import pyamr
from pyamr.core.regression.wregression import BaseRegressionWrapper
from pyamr.core.regression.wls import WLSWrapper
from pyamr.core.regression.arima import ARIMAWrapper
from pyamr.core.regression.sarimax import SARIMAXWrapper


class WLSARMAWrapper(BaseRegressionWrapper):
    # Attributes.
    _name = 'WLSARMA'  # Label to add to the attributes when saving.

    def _identifier(self, **kwargs):
        """This methods describes de model."""
        # Returns description
        return "%s + %s" % (self._wls._identifier(), self._arma._identifier())

    # --------------------------------------------------------------------------
    #                            SET VARIABLES
    # --------------------------------------------------------------------------
    def _params_from_summary(self):
        """Gets parameters from the summary result of the raw object.
        """
        # Return
        return {}

    def _init_result(self, alpha=0.05):
        """This method set all the variables into this class.

        Notes:
          - if instead of having the attribute series it is desired to set
          each element of the series as an attribute, just used the following
          statement: setattr(self, name, value).

          - Note the difference between the resid and wresid since they will
          also have different statistical properties. Furthermore, there is
          a vector with normalized residuals resid_pearson.

        @see: statsmodels.WLS
        @see: statsmodels.tsa.ARIMA
        @see: statsmodels.RegressionResultsWrapper

        Parameters
        ----------

        Returns
        -------
        """
        # Libraries.
        from statsmodels.iolib.summary import _getnames

        # Create series.
        d = {}

        # Add generic metrics.
        # Create params information.
        # Add coefficients statistics.

        # Further statistics.
        d.update(self._params_from_summary())
        d.update(self._resid_stats())

        # Return
        return d

    # --------------------------------------------------------------------------
    #                           HELPER METHODS
    # --------------------------------------------------------------------------
    def as_summary(self, **kwargs):
        """This method displays the summary.
        """
        # Variables.
        om, dw = 0.0, self.m_dw,
        pom, jb = 0.0, self.m_jb_value
        skew, pjb = self.m_skew, self.m_jb_prob
        kurt, cond = self.m_kurtosis, 0.0
        norm, pnorm = self.m_nm_value, self.m_nm_prob

        # Create summary.
        smry = "WLSARMA Regression Results".center(78)
        smry += "\n" + "=" * 78 + "\n"
        smry += "Omnibus:       %#21.3f  Durbin-Watson:    %#22.3f\n" % (om, dw)
        smry += "Prob(Omnibus): %#21.3f  Jarque-Bera (JB): %#22.3f\n" % (pom, jb)
        smry += "Skew:          %#21.3f  Prob(JB):         %#22.3f\n" % (skew, pjb)
        smry += "Kurtosis:      %#21.3f  Cond. No.:        %#22.3f\n" % (kurt, cond)
        smry += "Normal (N):    %#21.3f  Prob(N):          %#22.3f\n" % (norm, pnorm)
        smry += "=" * 78

        # Return
        return smry

    # --------------------------------------------------------------------------
    #                                 FIT
    # --------------------------------------------------------------------------
    def fit(self, endog, exog=None, weights=None, wls_kwargs={},
            arima_kwargs=None,
            sarimax_kwargs=None,
            **kwargs):
        """This method...

        Parameters
        ----------
        endog        :
        exog         :
        wls_kwargs   :
        arima_kwargs :
        sarimax_kwargs :
        kwargs :

        Returns
        -------
        """
        # Create weighted linear regression
        self._wls = WLSWrapper(sm.WLS).fit(endog=endog, exog=exog,
                                     weights=weights,
                                     **wls_kwargs)

        # Create best fitting for ARIMA.
        if arima_kwargs is not None:
            self._arma = \
                ARIMAWrapper(ARIMA).fit(endog=self._wls._resid, **arima_kwargs)

        # Create best fitting for SARIMAX.
        if arima_kwargs is None and sarimax_kwargs is not None:
            self._arma = \
                SARIMAXWrapper(SARIMAX).fit(endog=self._wls._resid, **sarimax_kwargs)

        # Set fitted values (what if arima d=1).
        self._fittedvalues = self.get_prediction()[1, :]

        # Set residuals
        self._resid = endog - self._fittedvalues

        # Set series.
        self._result = self._init_result(alpha=0.05)

        # Return
        return self

    def get_prediction(self, start=None, end=None, ptype='combined', **kwargs):
        """This method computes the prediction.

        Note: Correct for ARIMA. The arima is trained with n samples (e.g. 24)
        and then we ask for start=50 and end=74 hance for some reason arima
        becomes crazy and return a larger array. Therefore, we predict for
        indices None (begining) till the maxium number to predict which is
        fshape-start.

        Parameters
        ----------
        start : start of prediction (as in wls)
        end   : end of preediction (as in wls)
        type  : string (combined, wls, arma)

        Returns
        -------
        """
        # Create end for arma.
        end_arma = None if start is None or end is None else end - start

        # Return partial predictions.
        if ptype == 'wls':
            return self._wls.get_prediction(start=start, end=end)
        if ptype == 'arma':
            return self._arma.get_prediction(start=None, end=end_arma, **kwargs)

        # Compute both predictions
        pred_wls = self._wls.get_prediction(start=start, end=end)
        pred_arma = self._arma.get_prediction(start=None, end=end_arma, **kwargs)

        # Create combined predictions (careful with this).
        pred = np.empty(pred_wls.shape)
        pred[1, :] = pred_wls[1, :]
        pred[1, pred_arma[0, :].astype(int)] += pred_arma[1, :]

        # Return
        return pred


if __name__ == '__main__':
    # Import class.
    print("E")


##############################################################################
# Author: Bernard Hernandez
# Filename: 03-main-create-sari-idxs.py
# Description : This file contains differnent statistics used in time-series.
#               What it mainly does is to format the output of tests provided
#               by external libraries and return them in a dataframe.
#
# TODO: Move it to a module.
#
###############################################################################
# Forces decimals on divisions.
from __future__ import division 

# Libraries
import sys
import numpy as np
import pandas as pd

# Import base wrapper
from pyamr.core.stats.wbase import BaseWrapper


# ----------------------------------------------------------------------------
#
# ----------------------------------------------------------------------------

def kendall(x):  
  """   
  Parameters
  ----------
  x     : a vector of data
  alpha : significance level (0.05 default)

  Returns
  -------
  trend : tells the trend (increasing, decreasing or no trend)
  h     : True (if trend is present) or False (if trend is absence)
  p     : p value of the significance test
  z     : normalized test statistics 
  """
  # Libraries.
  from scipy.stats import norm

  # Compute n.
  n = len(x)

  # calculate S 
  s = 0
  for k in range(n-1):
      for j in range(k+1,n):
          s += np.sign(x[j] - x[k])

  # calculate the unique data
  unique_x = np.unique(x)
  g = len(unique_x)

  # calculate the var(s)
  if n == g: # there is no tie
      var_s = (n*(n-1)*(2*n+5))/18
  else: # there are some ties in data
      tp = np.zeros(unique_x.shape)
      for i in range(len(unique_x)):
          tp[i] = sum(unique_x[i] == x)
      var_s = (n*(n-1)*(2*n+5) + np.sum(tp*(tp-1)*(2*tp+5)))/18

  # calculate z
  if s>0:      z = (s - 1)/np.sqrt(var_s)
  elif s == 0: z = 0
  elif s<0:    z = (s + 1)/np.sqrt(var_s)

  # Check
  if np.isnan(s): return [None, None]

  # calculate the p_value
  p = 2*(1-norm.cdf(abs(z))) # two tail test

  # Return
  return [p, z]

class KendallWrapper(BaseWrapper):
  """
  """

  # --------------------------------------------------------------------------
  #                             new methods
  # --------------------------------------------------------------------------
  def trend_exists(self, alpha):
    """This method returns a boolean with the stationarity outocme.

    Parameters
    ----------
    alpha : float
      The significance level

    Returns
    -------
    """
    # Libraries.
    from scipy.stats import norm
    # Check.
    if self._raw[1] is None: return None
    # Return
    return abs(self._raw[1]) > norm.ppf(1-alpha/2) # abs(z)

  def trend_direction(self, alpha):
    """This method returns the trend direction.

    Parameters
    ----------
    alpha : float
      The significance level

    Returns
    -------
    """
    # Libraries.
    from scipy.stats import norm
    # Check
    if self._raw is None: return "failed"
    # Compute h
    h = abs(self._raw[1]) > norm.ppf(1-alpha/2) 
    # Trends
    z = self._raw[1]
    if (z<0) and h:   return 'decreasing'
    elif (z>0) and h: return 'increasing'
    else:             return 'no trend'

  # --------------------------------------------------------------------------
  #                           override methods
  # --------------------------------------------------------------------------
  def evaluate(self, alpha=0.05, **kwargs):
    """Evaluates the model.
    """
    # Create series
    d = {}

    # Add results
    d['m_pvalue'] = self._raw[0]
    d['m_z'] = self._raw[1]
    d['m_trend_existence'] = self.trend_exists(alpha)
    d['m_trend_direction'] = self.trend_direction(alpha)

    # Return
    return d
    
  def as_summary(self, alpha=0.05):
    """This method displays the summary.
    """
    # Create summary base
    summary = '     kendall test (monotonic)  \n'
    summary+= "==================================\n"
    summary+= "statistic (z):   %#17.3f\n" % self.m_z
    summary+= "pvalue (manual): %#17.5f\n" % self.m_pvalue
    summary+= "trend exists:    %17s\n" % self.trend_exists(alpha)
    summary+= "trend direction: %17s\n" % self.trend_direction(alpha)
    summary+= "=================================="
    # Return
    return summary








if __name__ == '__main__':
 
 
  # Libraries
  import pandas as pd

  # ----------------------------
  # set basic configuration
  # ----------------------------
  # Set pandas configuration.
  pd.set_option('display.max_colwidth', 14)
  pd.set_option('display.width', 150)
  pd.set_option('display.precision', 4)

  # ----------------------------
  # create data
  # ----------------------------
  # Constants
  length = 100
  offset = 100
  slope = 10

  # Create timeseries.
  x = np.arange(length)
  y = np.random.rand(length) * slope + offset

  # ---------------------
  # Create kendall object
  # ---------------------
  # Create object
  kendall = KendallWrapper(estimator=kendall).fit(x=y)

  # Print series.
  print("\n")
  print(kendall.as_series())

  # Print summary.
  print("\n")
  print(kendall.as_summary())

  # Print identifier
  print("\n")
  print(kendall._identifier())
 
  # -----------------
  # Save and load
  # -----------------
  # File location
  #fname = '../examples/saved/kendall-sample.pickle'

  # Save
  #kendall.save(fname=fname)

  # Load
  #kendall = KendallWrapper().load(fname=fname)